#
# Collective Knowledge
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#

# This module contains function to help writing unit tests.
# This module is automatically added to each unit test module during
# execution as the 'test_util' variable (much like the 'ck' variable).
# For examples of the usage of this functions, please looks at the kernel tests
# (ck/repo/kernel/test/test_*.py files).

import unittest
import sys
import os
from contextlib import contextmanager

ck = None  # Will be updated by CK (initialized CK kernel)

##############################################################################
# Temporary files helper.


@contextmanager
def tmp_file(suffix='', prefix='ck-test-', content=''):
    """
    Yields a temporary file name. If content is given (not empty), 
    creates the file with this content. Remove the file afterwards.
    """

    fname = ck.gen_tmp_file({'suffix': suffix, 'prefix': prefix})['file_name']
    try:
        if content != '':
            with open(fname, 'w') as f:
                f.write(content)
        yield fname
    finally:
        try:
            os.remove(fname)
        except OSError:
            pass

##############################################################################
# Temporary directory helper.


@contextmanager
def tmp_dir(suffix='', prefix='ck-test-', cwd=False):
    """
    Yields a temporary directory name (the directory is created). 

    If 'cwd' is True, make this directory the working directory. Restores the
    original working directory afterwards.

    Removes the directory and all its content afterwards.
    """

    import shutil
    import tempfile

    saved_cwd = os.getcwd()
    dname = tempfile.mkdtemp()
    try:
        if cwd:
            os.chdir(dname)
        yield dname
    finally:
        if cwd:
            os.chdir(saved_cwd)

        try:
            shutil.rmtree(dname)
        except OSError:
            pass

##############################################################################
# System streams helper.


@contextmanager
def tmp_sys(input_buf=''):
    """
    Yields nothing, but substitutes standard input, output and error streams with 
    StringIO streams. If 'input_buf' is given, make sure the new input stream
    contains it.

    Also substitutes the standard exit() routine with
    the one which simple prints exit code. 

    Restores the original streams and exit() afterwards.
    """

    saved_stdout = sys.stdout
    saved_stderr = sys.stderr
    saved_stdin = sys.stdin
    saved_exit = sys.exit
    try:
        out_stream = get_io()
        err_stream = get_io()
        in_stream = get_io(input_buf)

        sys.stdout = out_stream
        sys.stderr = err_stream
        sys.stdin = in_stream
        sys.exit = dummy_exit

        yield
    finally:
        sys.stdout = saved_stdout
        sys.stderr = saved_stderr
        sys.stdin = saved_stdin
        sys.exit = saved_exit

##############################################################################
# Temporary configuration helper.


@contextmanager
def tmp_cfg(cfg_key, cfg_value='yes'):
    """
    Yields nothing, but assignes the given configuration key (ck.cfg[cfg_key])
    the given value. Restores the old value afterwards.
    """

    saved_value = ck.cfg.get(cfg_key, None)
    try:
        ck.cfg[cfg_key] = cfg_value
        yield
    finally:
        if saved_value is None:
            ck.cfg.pop(cfg_key, None)
        else:
            ck.cfg[cfg_key] = saved_value

##############################################################################
# Temporary repository helper.


@contextmanager
def tmp_repo(name='ck-test-repo', cfg={}):
    """
    Creates a new repository and yields the information about it.

    Removes the repository afterwards.
    """

    d = {'module_uoa': 'repo', 'quiet': 'yes',
         'data_uoa': name, 'action': 'add'}
    d.update(cfg)
    r = ck.access(d)
    if 0 != r['return']:
        raise AssertionError(
            'Failed to create a temporary repo ' + name + ' : ' + r.get('error', None))
    try:
        path = r['dict']['path']
    except Exception:
        path = '<unknown>'
    try:
        yield r
    finally:
        r = ck.access({'module_uoa': 'repo', 'quiet': 'yes',
                       'data_uoa': name, 'action': 'remove'})
        if 0 != r['return']:
            raise AssertionError('Failed to remove temporary repo ' +
                                 name + ' at location ' + path + ' : ' + r.get('error', None))

##############################################################################
# Prints exit code to the system output. Used internally


def dummy_exit(code):
    print('Exit code: ' + str(code))

##############################################################################
# Returns StringIO buffer. Used internnaly


def get_io(buf=''):
    r = ck.string_io(buf)
    if not hasattr(r, 'encoding'):
        r.encoding = 'utf8'
    return r



#
# Collective Knowledge (script)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

import os


# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}


##############################################################################
# run a function in a Python script (main mode)

def run(i):
    """
    Input:  {
              data_uoa            - data UOA of the script entry
              (repo_uoa)          - repo UOA of the script entry
              (script_module_uoa) - module UOA of the script entry

              name                - subscript name (from entry desc - will be called via shell)
              (params)            - pass params to CMD

                or

              (code)              - Python script name (without .py)
              (func)              - Python function name in this script
              (dict)              - dict to pass to script
              (output_json_file)  - filename to save the output dictionary into

            }

    Output: {
              return        - return code =  0, if successful
                                          >  0, if error
              (error)       - error text if return > 0

              (return_code) - script's return code
            }

    """


    ruoa=i.get('repo_uoa','')
    muoa=i.get('script_module_uoa','')
    if muoa=='': muoa=work['self_module_uoa']
    duoa=i.get('data_uoa','')

    name=i.get('name','')
    params=i.get('params','')


    # Loading entry
    rx=ck.access({'action':'load',
                  'module_uoa': muoa,
                  'data_uoa':duoa})
    if rx['return']>0: return rx
    d=rx['dict']
    p=rx['path']

    python_script_name  = i.get('code', d.get('default_python_script_name') )
    function_name       = i.get('func', d.get('default_function_name') )

    if python_script_name and function_name:
       r=ck.load_module_from_path({'path':p, 'module_code_name':python_script_name, 'skip_init':'yes'})
       if r['return']>0: return r

       loaded_module=r.get('code')
       if not loaded_module:
          return {'return':1, 'error':'no python code found'}

       script_function=getattr(loaded_module, function_name)

       if not script_function:
          return {'return':1, 'error':'function '+function_name+' not found in python script '+python_script_name}

       func_input_data=i.get('dict',{})

       func_input_data['ck_kernel']=ck

       rr=script_function(func_input_data)
       if rr['return']>0:
            return {'return':1, 'error':'script failed ('+rr['error']+')'}      # LG: why not just return rr?

       rr['return_code']=rr['return']       # LG: for compatibility with a different protocol?

       output_json_file = i.get('output_json_file')
       if output_json_file:
            r=ck.save_json_to_file({'json_file': output_json_file, 'dict':rr})
            if r['return']>0: return r

    else:
       ss=d.get('sub_scripts',{})

       xs=ss.get(name,{})

       if len(xs)==0:
          return {'return':1, 'error':'subscript "'+name+'" is not found in entry "'+duoa+'"'}

       cmd=xs.get('cmd','')

       p1=p+os.path.sep

       cmd=cmd.replace('$#ck_path#$', p1)+' '+params

       rx=os.system(cmd)

       rr={'return':0, 'return_code':rx}

    return rr


##############################################################################
# run a function in a Python script via a CK wrapper

def _run_external(i):
    """
    Input:  {
                data_uoa            - data UOA of the script-containing entry
                (script_module_uoa) - module UOA of the script-containing entry ('script' by default)
                (prewrapper_lines)  - lines of the environment-setting pre-wrapper
                (script_module_uoa) - module UOA of the script entry
                (keep_tmp_files)    - "yes" will keep them, deleted otherwise

                (code)              - Python script name (without .py)
                (func)              - Python function name in this script
                (dict)              - dict to pass to script
                (output_json_file)  - filename to save the output dictionary into

            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0

                (return_code)       - script's return code
            }
    Test:
            ck _run_external script:debug @@@"{'prewrapper_lines': ['export CK_PYTHON=python3'], 'dict': {'alpha': 777}}"

    """

    module_uoa          = i.get('script_module_uoa', work['self_module_uoa'])
    data_uoa            = i['data_uoa']
    prewrapper_lines    = i.get('prewrapper_lines', [])
    keep_tmp_files      = i.get('keep_tmp_files', 'no')
    code                = i.get('code','')
    func                = i.get('func','')
    func_input_data     = i.get('dict', {})

    code = '--code=' + code if code else ''
    func = '--func=' + func if func else ''

    ## Generate one common seed for two temporary names for JSON files:
    #
    gentmp_adict = ck.gen_tmp_file({'prefix': 'ck_call_', 'suffix': '.json'})
    if gentmp_adict['return']>0: return gentmp_adict

    ## Keeping the second name in sync with the first one:
    #
    input_json_file = gentmp_adict['file_name']
    output_json_file = input_json_file.replace('ck_call_', 'ck_response_')

    ## Avoiding to serialize the kernel:
    #
    func_input_data.pop('ck_kernel')

    ## Fill in the input_json_file with input parameters:
    #
    input_adict = ck.save_json_to_file( {'json_file': input_json_file, 'dict': {'dict': func_input_data} } )
    if input_adict['return']>0: return input_adict

    ## Form and run the external ck command with the given Python:
    #
    shell_cmd = "ck run {}:{} {} {} @{} --output_json_file={}".format(module_uoa, data_uoa, code, func, input_json_file, output_json_file)

    ck.out("RPC call: {}".format(shell_cmd))
    envsecute_adict = ck.access({'action':'envsecute',
        'module_uoa':           'env',
        'prewrapper_lines':     prewrapper_lines,
        'shell_cmd':            shell_cmd,
    })
    ck.out("RPC returned: {}".format(envsecute_adict))
    return_code = envsecute_adict['return']

    if keep_tmp_files=='yes':
        ck.out("Temporary JSON files' generated: {} and {}".format(input_json_file, output_json_file))

    ## Parse the output_json_file and return its contents:
    #
    output_adict = ck.load_json_file( {'json_file': output_json_file} )
    if output_adict['return']>0: return output_adict

    output_dict = output_adict['dict']
    output_dict['return_code'] = return_code

    if return_code==0 and output_dict['return']==0 and keep_tmp_files!='yes':
        os.remove(input_json_file)
        os.remove(output_json_file)

    return output_dict



#  Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Example of DNNClassifier for Iris plant dataset.
This example uses APIs in Tensorflow 1.4 or above.
"""

# Converted by Grigori Fursin to the CK format (http://cKnowledge.org)
#   from https://github.com/tensorflow/tensorflow/blob/master/tensorflow/examples/learn/iris.py

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import os
import json
import numpy as np
import tensorflow as tf

ck_params='ck-params.json'

def main(i):

  mode=i['mode']
  fi=i['input_file']

  # Load input file
  with open(fi) as f:
    s=f.read()
    d = json.loads(s)
    f.close()

  ftable=d['ftable']

  ctable=d.get('ctable',[])
  model_params=d.get('model_params',{})

  fo=d['output_file']
  fod=d['model_dir']

  # Prepare model parametrs
  if mode=='train':
     # Check distinct labels
     labels=[]
     max_label=0
     for q1 in ctable:
         q=q1[0]
         if q not in labels:
            labels.append(q)
            if q>max_label: max_label=q
#     xn_classes=len(labels)
     xn_classes=max_label+1

     xhidden_units=model_params.get('hidden_units',[])
     if len(xhidden_units)==0: xhidden_units=[10, 20, 10]

     feature_length=len(ftable[0])
  else:
     # Read ck-params.json
     x=os.path.join(fod, ck_params)

     with open(x) as f:
       s=f.read()
       dx = json.loads(s)
       f.close()

     xn_classes=dx['n_classes']
     feature_length=dx['feature_length']
     xhidden_units=dx['hidden_units']

  # Prepare model
  # Specify that all features have real-value data
  feature_columns = [tf.feature_column.numeric_column("x", shape=[feature_length])]

  classifier = tf.estimator.DNNLinearCombinedClassifier(
                                          linear_feature_columns=feature_columns,
                                          dnn_feature_columns=feature_columns,
                                          dnn_hidden_units=xhidden_units,
                                          n_classes=xn_classes,
                                          model_dir=fod)

  # Use model
  if mode=='train':
     # Train model.
     print ('')
     print ('Training ...')
     print ('')

     xsteps=model_params.get('training_steps','')
     if xsteps=='' or xsteps==None: xsteps="2000"
     xsteps=int(xsteps)

     train_input_fn = tf.estimator.inputs.numpy_input_fn(
         x={"x": np.array(ftable)},
         y=np.array(ctable),
         num_epochs=None,
         shuffle=True)

     classifier.train(input_fn=train_input_fn, steps=xsteps)

     ftable_test=d.get('ftable_test',[])
     if len(ftable_test)==0: ftable_test=ftable

     ctable_test=d.get('ctable_test',[])
     if len(ctable_test)==0: ctable_test=ctable

     # Define the test inputs
     print ('')
     print ('Testing ...')
     print ('')

     test_input_fn = tf.estimator.inputs.numpy_input_fn(
         x={"x": np.array(ftable_test)},
         y=np.array(ctable_test),
         num_epochs=1,
         shuffle=False)

     # Evaluate accuracy.
     accuracy_score = classifier.evaluate(input_fn=test_input_fn)["accuracy"]

     print ('')
     print ('Test Accuracy: {0:f}'.format(accuracy_score))
     print ('')

     dd={'output_dir':fod,
         'accuracy':float(accuracy_score),
         'hidden_units':xhidden_units,
         'n_classes':xn_classes,
         'feature_length':feature_length}

     # Record model info
     s=json.dumps(dd,indent=2,sort_keys=True)
     with open(fo,'w') as f:
        f.write(s)
        f.close()

     # Record model info to TF model dir
     x=os.path.join(fod, ck_params)
     with open(x,'w') as f:
        f.write(s)
        f.close()

  ##############################################################################
  elif mode=='prediction':
     # Classify samples
     predict_input_fn = tf.estimator.inputs.numpy_input_fn(
         x={"x": np.array(ftable, dtype=np.float32)},
         num_epochs=1,
         shuffle=False)

     ctable=[]

     print ('')
     print ('Predictions:')

     predictions = list(classifier.predict(input_fn=predict_input_fn))

#     predictions1 = np.squeeze(predictions) # FGG: don't need it - wrong when only one entry

     for q in range(0, len(predictions)):
         print (str(q)+') '+str(np.asscalar(predictions[q]['class_ids'][0])))
         ctable.append(int(np.asscalar(predictions[q]['class_ids'][0])))

     # Record prediction
     dd={'ftable':ftable,
         'ctable':ctable}

     print ('')
     print ('Recording results to '+fo+' ...')
     print ('')

     s=json.dumps(dd,indent=2,sort_keys=True)
     with open(fo,'w') as f:
        f.write(s)
        f.close()

  else:
     print ('Error in CK-TF wrapper: mode "'+mode+'" is not supported ...')
     exit(1)

  return

if __name__ == "__main__":
    argv=sys.argv[1:]

    if len(argv)<2:
       print ('Not enough command line arguments ...')
       exit(1)

    mode=argv[0]
    input_file=argv[1]

    main({'mode':mode, 'input_file':input_file})



#
# Collective Knowledge (Jupyter Notebook)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# remote output from Jupyter Notebook

def clean(i):
    """
    Input:  {
              file_in    - input jupyter notebook file
              (file_out) - output jupyter notebook file (otherwise {in}.out
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o=i.get('out','')

    fin=i.get('file_in','')
    if fin=='':
       return {'return':1, 'error':'Usage: ck clean jnotebook --in={juypter notebook file} (--out={output file})'}

    fout=i.get('file_out','')
    if fout=='':
       fout=fin+'.out'

    r=ck.load_json_file({'json_file':fin})
    if r['return']>0: return r

    d=r['dict']

    cells=d['cells']

    for ic in range(0,len(cells)):
        c=cells[ic]
        if 'outputs' in c:
           c['outputs']=[]
        cells[ic]=c

    r=ck.save_json_to_file({'json_file':fout, 'dict':d})
    if r['return']>0: return r

    if o=='con':
       ck.out('Output file: '+fout)

    return {'return':0}

##############################################################################
# run Jupyter Notebook from a CK entry

def run(i):
    """
    Input:  {
              data_uoa     - CK Jupyter notebook entry
              (name)       - full name of file (if more than one)

              (original)   - if 'yes', do not generate tmp file
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import shutil

    duoa=i.get('data_uoa','')
    if duoa=='':
       return {'return':1, 'error':'Usage: ck run jnotebook:{UOA} (--name={notebook filename if more than one}'}

    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duoa})
    if r['return']>0: return r

    p=r['path']

    name=i.get('name','')
    if name=='':
       ld=os.listdir(p)
       nbs=[]
       for f in ld:
           if f.endswith('.ipynb'):
              nbs.append(f)
       if len(nbs)==0:
          return {'return':1, 'error':'can\'t find \ipython/jupyter notebooks in the CK entry'}
       name=nbs[0]

    # Check if need tmp file or not
    ff=os.path.join(p,name)
    if i.get('original','')!='yes':
       rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':'.ipynb', 'remove_dir':'no'})
       if rx['return']>0: return rx
       ftmp=rx['file_name']

       shutil.copy(ff, ftmp)
       ff=ftmp

    cmd='jupyter notebook '+ff
    os.system(cmd)

    return {'return':0}



#
# Collective Knowledge: compiler flags crowdtuning (crowdsource autotuning via spare computers such as mobile devices)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module
import copy

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# crowdsource these experiments

def crowdsource(i):
    """
    See in module "experiment.tune.compiler.flags"
    """

    i['module_uoa']='experiment.tune.compiler.flags'
    i['skip_pruning']='yes'
    i['module_cfg']=copy.deepcopy(cfg)
    i['module_work']=copy.deepcopy(work)
    return ck.access(i)

##############################################################################
# view solutions in html

def html_viewer(i):
    """
    See in module "experiment.tune.compiler.flags"
    """

    i['module_uoa']='experiment.tune.compiler.flags'
    i['module_cfg']=copy.deepcopy(cfg)
    i['module_work']=copy.deepcopy(work)
    return ck.access(i)

##############################################################################
# replay optimization

def replay(i):
    """
    See in module "program.optimization"
    """

    i['module_uoa']=cfg['module_deps']['program.optimization']
    i['module_ref_uoa']=work['self_module_uid']
    i['module_cfg']=copy.deepcopy(cfg)
    i['module_work']=copy.deepcopy(work)
    return ck.access(i)

##############################################################################
# prune compiler flags to find minimal set of choices

def prune(i):
    """
    See in module "program.optimization"
    """

    i['module_uoa']=cfg['module_deps']['program.optimization']
    i['module_ref_uoa']=work['self_module_uid']
    i['module_cfg']=copy.deepcopy(cfg)
    i['module_work']=copy.deepcopy(work)
    return ck.access(i)



#
# Copyright (c) 2019 cTuning foundation.
# See CK COPYRIGHT.txt for copyright details.
#
# SPDX-License-Identifier: BSD-3-Clause.
# See CK LICENSE.txt for licensing details.
#
# Collective Knowledge - raw data access (JSON).
#
# Developers:
# - Nikolay Istomin, Xored.
# - Anton Lokhmotov, dividiti.
# - Leo Gordon, dividiti.
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel)

import os
import sys
import json
import re
import traceback
from collections import defaultdict
from pprint import pprint

import pandas as pd
import numpy as np

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}


##############################################################################
# Get raw data for visualization widget.

def get_raw_data(i):
    """
    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }
    """
    prefilter_mode = i.get('prefilter_mode', 'image_classification_singlestream')
    prefilter_config = cfg['prefilter_config'][prefilter_mode]

    def get_experimental_results_from_cache(cache_dir=None, cache_file=None):
        if cache_dir is None:
            # Check this module's directory.
            cache_repo_uoa    = 'ck-mlperf'
            cache_module_uoa  = 'module'
            cache_data_uoa    = 'mlperf.inference'
            r = ck.access({'action':'find',
                'repo_uoa':cache_repo_uoa, 'module_uoa':cache_module_uoa, 'data_uoa':cache_data_uoa})
            if r['return']>0:
                ck.out('Error: %s' % r['error'])
                exit(1)
            cache_dir = r['path']

        if os.path.exists(cache_dir):
            ck.out("Loading the results table(s) from cache at '%s' ..." % cache_dir)
            if cache_file is None:
                # Check all zip files in the cache directory.
                dfs = []
                for cache_file in os.listdir(cache_dir):
                    if cache_file.endswith(".zip"):
                        ck.out("- '%s'" % cache_file)
                        cache_path = os.path.join(cache_dir, cache_file)
                        df = pd.read_pickle(cache_path)
                        df['Cache'] = cache_file
                        df.set_index('Cache', drop=True, append=True, inplace=True)
                        dfs.append(df)
                df = pd.concat(dfs)
            else:
                cache_path = os.path.join(cache_dir, cache_file)
                df = pd.read_pickle(cache_path)
        else:
            ck.out("No cache at '%s' ..." % cache_dir)
            df = pd.DataFrame()
        return df

    def df_as_record(df):
        for index, record in df.to_dict(orient='index').items():
            indices = index if isinstance(index, list) else [ index ]
            record.update({ n:v for n,v in zip(df.index.names, indices) })
            yield record

    def to_value(i):
        if type(i) is str and i.startswith('http'):
            return '<a href="%s">Link</a>' % i

        if type(i) is np.ndarray:
            return i.tolist()

        if isinstance(i, np.int64):
            return int(i)

        if isinstance(i, np.float64):
            return float(i)

        return i

    # Only the official v0.5 results which are cached in this module's directory.
#    df = get_experimental_results_from_cache(cache_file='mlperf-inference-v0.5-results.zip')
#    df = df.set_index('ID', drop=True)
    # All the results in this module's directory.
    df = get_experimental_results_from_cache(cache_file=None)
    
    debug_output = i.get('out')=='con'
    table = []
    for record in df_as_record(df):
        row = {}
        props = [
            'ID',
            'Division',
            'Category',
            'Submitter',
            'System',
            'Task',
            'Benchmark',
            'Scenario',
            'Processor',
            'Processor #',
            'Accelerator',
            'Accelerator #',
#            'Software',
            'FF_M',
            'FF_E',
            'FF_D',
            'FF_S',
            'Details',
            'Code',
            'Notes'
        ]
        for prop in props:
            row[prop] = to_value(record.get(prop, ''))
        row['Software'] = { 'title': 'Software', 'cmd':  record.get('Software', '') }
        row['Invalid?'] = len([ note for note in record.get('Notes', '') if str(note) == 'Result = INVALID' ])

        for score in prefilter_config['score_columns']:
            score_no_scenario = score[0:5] # e.g. A_NMT
            row[score_no_scenario] = np.nan_to_num(record.get(score, 0.0))

        table.append(row)
        if debug_output:
            ck.out(str(row))

    merged_table = table

    return { 'return': 0, 'table': merged_table }


##############################################################################
# Get raw config for visualization widget.

def get_raw_config(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    data_config = cfg['data_config']
    data_config['return'] = 0

    return data_config



#
# Collective Knowledge
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#

import unittest
import os
import shutil
import tempfile
import sys
from contextlib import contextmanager

ck = None           # Will be updated by CK (initialized CK kernel)
test_util = None    # Will be updated by CK (initialized CK test utils)


def merge(dict1, dict2):
    ret = dict1.copy()
    ret.update(dict2)
    return ret

# Tests for repo/module/entry management operations like creating, adding actions, deleting, etc.
# Generally, these are more complex tests, then those from 'test_kernel.py'. These tests
# may take some time to execute, they may require Internet connection, etc.


class TestMgmt(unittest.TestCase):

    def test_entry_crud(self):
        entry = {'module_uoa': 'kernel',
                 'repo_uoa': 'local', 'data_uoa': 'ck-test-abcd'}
        r = ck.add(entry)
        self.assertEqual(0, r['return'], r.get('error', None))
        try:
            self.assertIn('data_uid', r)
            self.assertIn('data_name', r)
            self.assertIn('path_module', r)

            path = r['path']
            uid = r['data_uid']

            self.assertEqual('ck-test-abcd', r['data_uoa'])
            self.assertEqual('ck-test-abcd', r['data_name'])
            self.assertEqual('local', r['repo_uoa'])

            self.assertEqual({}, ck.load_meta_from_path(
                {'path': path})['dict'])

            module_path = os.path.join(ck.find_path_to_repo(
                {'repo_uoa': 'local'})['path'], 'kernel')
            r = ck.find_path_to_entry({'path': module_path, 'data_uoa': uid})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(path, r['path'])

            missing_data = ck.gen_uid({})['data_uid']
            r = ck.find_path_to_entry(
                {'path': module_path, 'data_uoa': missing_data})
            self.assertEqual(-1, r['return'])

            meta_dict = {'a': 1, 'b': [2, 3]}
            r = ck.update(merge(entry, {'dict': meta_dict}))
            self.assertEqual(0, r['return'], r.get('error', None))

            self.assertEqual(meta_dict, ck.load_meta_from_path(
                {'path': path})['dict'])

            r = ck.rename(merge(entry, {'new_data_uoa': 'ck-test-0123'}))
            self.assertEqual(0, r['return'], r.get('error', None))

            self.assertNotEqual(
                meta_dict, ck.load_meta_from_path({'path': path})['return'])

            # rename back
            r = ck.rename(
                merge(entry, {'data_uoa': 'ck-test-0123', 'new_data_uoa': 'ck-test-abcd'}))
            self.assertEqual(0, r['return'], r.get('error', None))

            r = ck.copy(merge(entry, {'new_data_uoa': 'ck-test-abcd-copy'}))
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(path + '-copy', r['path'])

            r = ck.move(merge(entry, {'data_uoa': 'ck-test-abcd-copy',
                                      'new_module_uoa': 'test', 'new_data_uoa': 'ck-test-abcd'}))
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertNotEqual(0, ck.load_meta_from_path(
                {'path': path + '-copy'})['return'])
            self.assertEqual(meta_dict, ck.load_meta_from_path(
                {'path': r['path']})['dict'])

            r = ck.delete(merge(entry, {'module_uoa': 'test'}))
            self.assertEqual(0, r['return'], r.get('error', None))

            fname = path + '/test.txt'
            with open(fname, 'w') as f:
                f.write('test\n')
            self.assertTrue(os.path.isfile(fname))

            r = ck.delete_file(merge(entry, {'filename': 'test.txt'}))
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertFalse(os.path.isfile(fname))

        finally:
            r = ck.remove(entry)
            self.assertEqual(0, r['return'], r.get('error', None))

    def test_create_entry(self):
        with test_util.tmp_repo() as r:
            path = r['path']
            r = ck.create_entry({'path': path, 'force': 'yes'})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(os.path.join(path, r['data_uid']), r['path'])
            self.assertEqual('', r['data_alias'])
            self.assertEqual(r['data_uoa'], r['data_uid'])

            uid = ck.gen_uid({})['data_uid']
            r = ck.create_entry({'path': path, 'data_uid': uid})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(os.path.join(path, r['data_uid']), r['path'])
            self.assertEqual(uid, r['data_uid'])
            self.assertEqual('', r['data_alias'])
            self.assertEqual(r['data_uoa'], r['data_uid'])

            # second time with the same UID should fail
            r = ck.create_entry({'path': path, 'data_uid': uid})
            self.assertEqual(16, r['return'])

            # uid instead of uoa
            uid = ck.gen_uid({})['data_uid']
            r = ck.create_entry({'path': path, 'data_uoa': uid})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(os.path.join(path, r['data_uid']), r['path'])
            self.assertEqual(uid, r['data_uid'])
            self.assertEqual('', r['data_alias'])
            self.assertEqual(r['data_uoa'], r['data_uid'])

            # with alias
            alias = 'create-entry-test'
            r = ck.create_entry({'path': path, 'data_uoa': alias})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(os.path.join(path, alias), r['path'])
            self.assertIn('data_uid', r)
            self.assertEqual(alias, r['data_alias'])
            self.assertEqual(alias, r['data_uoa'])

    def test_lock(self):
        with test_util.tmp_repo() as repo:
            import time

            path = repo['path']

            r = ck.set_lock({'path': path, 'get_lock': 'yes'})
            self.assertEqual(0, r['return'], r.get('error', None))

            lock_uid = r['lock_uid']

            r = ck.set_lock({'path': path, 'get_lock': 'yes',
                             'lock_retries': '1', 'lock_retry_delay': '1'})
            self.assertEqual(32, r['return'])

            r = ck.set_lock({'path': path, 'unlock_uid': lock_uid+'1'})
            self.assertEqual(32, r['return'])
            self.assertEqual('entry is locked with another UID', r['error'])

            r = ck.set_lock({'path': path, 'unlock_uid': lock_uid})
            self.assertEqual(0, r['return'], r.get('error', None))

            r = ck.set_lock({'path': path, 'get_lock': 'yes',
                             'lock_expire_time': '1', 'unlock_uid': lock_uid})
            self.assertEqual(0, r['return'], r.get('error', None))
            time.sleep(1)

            r = ck.set_lock(
                {'path': path, 'lock_expire_time': '1', 'get_lock': 'yes'})
            self.assertEqual(0, r['return'], r.get('error', None))
            lock_uid = r['lock_uid']

            r = ck.check_lock({'path': path})
            self.assertEqual(32, r['return'])
            self.assertEqual('entry is locked', r['error'])

            r = ck.check_lock({'path': path, 'unlock_uid': lock_uid + '1'})
            self.assertEqual(32, r['return'])
            self.assertEqual('entry is locked with different UID', r['error'])

            time.sleep(2)
            r = ck.check_lock({'path': path, 'unlock_uid': lock_uid + '1'})
            self.assertEqual(32, r['return'])
            self.assertEqual('entry lock UID is not matching', r['error'])

            r = ck.check_lock({'path': path, 'unlock_uid': lock_uid})
            self.assertEqual(0, r['return'], r.get('error', None))

            r = ck.check_lock({'path': path, 'unlock_uid': lock_uid})
            self.assertEqual(32, r['return'])
            self.assertEqual('lock was removed or expired', r['error'])

    def test_perform_action_on_remote_repo(self):
        with test_util.tmp_repo(cfg={'remote': 'yes', 'url': 'http://cknowledge.org/repo/ck.php?'}) as repo:
            repo_uoa = repo['data_uid']
            with test_util.tmp_sys():
                r = ck.perform_action({
                    'repo_uoa': repo_uoa,
                    'data_uoa': 'default',
                    'module_uoa': 'repo',
                    'action': 'search',
                    'out': 'con'})
                self.assertEqual(0, r['return'], r.get('error', None))
                self.assertEqual('default:repo:default',
                                 sys.stdout.getvalue().strip())

        with test_util.tmp_repo(cfg={'remote': 'yes'}) as repo:
            repo_uoa = repo['data_uid']
            with test_util.tmp_sys():
                r = ck.perform_action(
                    {'repo_uoa': repo_uoa, 'action': 'search'})
                self.assertEqual(1, r['return'])
                self.assertEqual(
                    'URL of remote repository is not defined', r['error'])

    def test_perform_remote_action(self):
        r = ck.perform_remote_action({
            'module_uoa': 'kernel',
            'remote_server_url': 'http://cknowledge.org/repo/ck.php?',
            'data_uoa': 'default',
            'repo_uoa': 'default',
            'action': 'search',
            'cid': 'invalid_cid_for_deletion'})
        self.assertEqual(0, r['return'], r.get('error', None))
        lst = r['lst']
        self.assertEqual(1, len(lst))
        r = lst[0]
        self.assertEqual('kernel', r['module_uoa'])
        self.assertEqual('default', r['data_uoa'])
        self.assertEqual('default', r['repo_uoa'])
        self.assertIn('repo_uid', r)
        self.assertIn('module_uid', r)
        self.assertIn('data_uid', r)
        self.assertIn('path', r)

        with test_util.tmp_sys():
            r = ck.perform_remote_action({
                'module_uoa': 'kernel',
                'remote_server_url': 'http://cknowledge.org/repo/ck.php?',
                'data_uoa': 'default',
                'repo_uoa': 'default',
                'action': 'search',
                'out': 'con'})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual('default:kernel:default',
                             sys.stdout.getvalue().strip())

        r = ck.perform_remote_action({
            'module_uoa': 'kernel',
            'remote_server_url': 'http://cknowledge.org/repo/ck.php?',
            'data_uoa': 'default',
            'repo_uoa': 'default',
            'action': 'push',
            'cids': ['']})
        self.assertEqual(1, r['return'])
        self.assertEqual('filename is empty', r['error'])

        missing_filename = ck.gen_uid({})['data_uid']
        r = ck.perform_remote_action({
            'module_uoa': 'kernel',
            'remote_server_url': 'http://cknowledge.org/repo/ck.php?',
            'data_uoa': 'default',
            'repo_uoa': 'default',
            'action': 'push',
            'filename': missing_filename})
        self.assertEqual(1, r['return'])
        self.assertEqual('file '+missing_filename+' not found', r['error'])

        with test_util.tmp_file(content='abcd') as fname:
            missing_user = ck.gen_uid({})['data_uid']
            missing_pass = ck.gen_uid({})['data_uid']
            r = ck.perform_remote_action({
                'module_uoa': 'kernel',
                'remote_server_url': 'http://cknowledge.org/repo/ck.php?',
                'data_uoa': 'default',
                'repo_uoa': 'default',
                'action': 'push',
                'filename': fname,
                'remote_server_user': missing_user,
                'remote_server_pass': missing_pass})
            self.assertEqual(1, r['return'])
            self.assertEqual(
                'CK error: [kernel] writing to default repo is forbidden!', r['error'].strip())

        with test_util.tmp_dir(cwd=True):
            r = ck.perform_remote_action({
                'module_uoa': 'kernel',
                'remote_server_url': 'http://cknowledge.org/repo/ck.php?',
                'data_uoa': 'default',
                'repo_uoa': 'default',
                'action': 'pull',
                'filename': '.cm/meta.json',
                'out': 'con'})
            self.assertEqual(0, r['return'], r.get('error', None))

    def test_actions_crud(self):
        entry = {
            'module_uoa': 'module',
            'repo_uoa': 'local',
            'data_uoa': 'ck-test-abcd',
            'common_func': 'yes',
            'action': 'add',
            'dict': {'license': '1', 'copyright': '2', 'developer_email': '3', 'actions': {}, 'developer_webpage': '4', 'developer': '5', 'desc': 'abcd'},
            'sort_keys': 'yes'
        }
        r = ck.add(entry)
        self.assertEqual(0, r['return'], r.get('error', None))
        try:
            r = ck.add_action(merge(entry, {
                              'func': 'test_func', 'desc': 'test descr', 'skip_appending_dummy_code': 'yes'}))
            self.assertEqual(0, r['return'], r.get('error', None))

            r = ck.list_actions(entry)
            self.assertEqual(0, r['return'], r.get('error', None))
            lst = r['actions']
            self.assertEqual(1, len(lst))
            self.assertEqual('test descr', lst['test_func']['desc'])

            r = ck.remove_action(merge(entry, {'func': 'test_func'}))
            self.assertEqual(0, r['return'], r.get('error', None))

            r = ck.list_actions(entry)
            self.assertEqual(0, r['return'], r.get('error', None))
            lst = r['actions']
            self.assertEqual(0, len(lst))
        finally:
            r = ck.remove(entry)
            self.assertEqual(0, r['return'], r.get('error', None))

    def test_zip_unzip(self):
        with test_util.tmp_file() as fname:
            r = ck.zip({'action': 'zip', 'repo_uoa': 'default',
                        'module_uoa': 'module', 'data_uoa': 'repo', 'archive_name': fname})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertTrue(os.path.isfile(fname))

            dirname = tempfile.mkdtemp()
            self.assertEqual([], os.listdir(dirname))
            try:
                r = ck.unzip_file({'archive_file': fname, 'path': dirname})
                files = os.listdir(dirname)
                self.assertNotEqual([], files)
                self.assertIn('kernel', files)
                self.assertIn('module', files)
                self.assertIn('repo', files)
                self.assertIn('test', files)
            finally:
                shutil.rmtree(dirname)



#
# Collective Knowledge (module providing API to communicate with various ck-crowdnodes ...)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
import os

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# push file to ck-crowdnode

def push(i):
    """
    Input:  {
              (url)      - URL to ck-crowdnode (http://localhost:3333 by default)
              (keyfile)  - path to key

              (filename)  - local file to push to crowd-node
              (filename2) - file with relative path (relative path will be extracted 
                            to 'extra_path') - needed for compatibility with ADB/SSH

              (extra_path)          - extra path inside entry (create if doesn't exist)
              (archive)             - if 'yes' push to entry and unzip ...
              (overwrite)           - if 'yes', overwrite files
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o=i.get('out','')

    url=i.get('url','')
    if url=='': url='http://localhost:3333'

    keyfile=i.get('keyfile','')
    key=''
    if keyfile!='':
        # Load keyfile
        r=ck.load_text_file({'text_file':keyfile})
        if r['return']>0: return r
        key=r['string'].strip()

    fn=i.get('filename','')
    if fn=='' or not os.path.isfile(fn):
        return {'return':1, 'error':'file '+fn+' not found'}

    ep=i.get('extra_path','')
    if ep=='':
        fn2=i.get('filename2','')
        if fn2!='':
            ep=os.path.dirname(fn2)

    ii={'action':'push',
        'remote_server_url':url,
        'secretkey':key,
        'extra_path':ep,
        'filename':fn}
    return ck.access(ii)

##############################################################################
# pull file from ck-crowdnode

def pull(i):
    """
    Input:  {
              (url)      - URL to ck-crowdnode (http://localhost:3333 by default)
              (keyfile)  - path to key

              (filename)  - file to pull from remote crowd-node
              (filename2) - local file to record (to current path)

              (extra_path)          - extra path inside entry (create if doesn't exist)
              (archive)             - if 'yes' push to entry and unzip ...
              (overwrite)           - if 'yes', overwrite files
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o=i.get('out','')

    url=i.get('url','')
    if url=='': url='http://localhost:3333'

    keyfile=i.get('keyfile','')
    key=''
    if keyfile!='':
        # Load keyfile
        r=ck.load_text_file({'text_file':keyfile})
        if r['return']>0: return r
        key=r['string'].strip()

    fn=i.get('filename','')
    if fn=='':
        return {'return':1, 'error':'remote file not specified'}

    ep=os.path.dirname(fn)
    fn=os.path.basename(fn)

    ii={'action':'pull',
        'remote_server_url':url,
        'secretkey':key,
        'filename':fn,
        'extra_path':ep}
    r=ck.access(ii)
    if r['return']>0: return r

    return {'return':0}

##############################################################################
# execute command on ck-crowdnode

def shell(i):
    """
    Input:  {
              (url)      - URL to ck-crowdnode (http://localhost:3333 by default)
              (keyfile)  - path to key

              (cmd)      - command line
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              stdout
              stderr
              return_code
            }

    """

    import base64

    o=i.get('out','')

    url=i.get('url','')
    if url=='': url='http://localhost:3333'

    keyfile=i.get('keyfile','')
    key=''
    if keyfile!='':
        # Load keyfile
        r=ck.load_text_file({'text_file':keyfile})
        if r['return']>0: return r
        key=r['string'].strip()

    cmd=i.get('cmd','').strip()
    if cmd=='':
        return {'return':1, 'error':'cmd is not specified'}

    ii={'action':'shell',
        'remote_server_url':url,
        'secretkey':key,
        'cmd':cmd,
        'out':''}
    r=ck.access(ii)
    if r['return']>0: return r

    rc=r.get('return_code','')
    if rc=='': rc='0'
    irc=int(rc)

    xso=str(r.get('stdout_base64',''))
    xse=str(r.get('stderr_base64',''))

    enc=r.get('encoding','')
    if enc=='': enc='UTF-8'

    so=''
    if xso!='':
        so=base64.urlsafe_b64decode(xso)
        if type(so)==bytes:
            so=so.decode(encoding=enc, errors='ignore')

    se=''
    if xse!='':
        se=base64.urlsafe_b64decode(xse)
        if type(se)==bytes:
            se=se.decode(encoding=enc, errors='ignore')

    if o=='con':
        if so!='': 
            ck.out(so)
        if se!='': 
            if getattr(ck, 'eout', None)==None:
                ck.out(se)
            else:
                ck.eout(se)

    # Check return code
    if irc>0:
        return {'return':irc, 'error':'remote command likely failed (return code > 0)'}

    return {'return':0, 'stdout':so, 'stderr':se}



#
# Collective Knowledge (Grigori's misc research functions)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}


##############################################################################
# Just a centralized place to get a home directory

def get_home_dir(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    from os.path import expanduser
    home_dir = expanduser("~")

    if i.get('out', '') == 'con':
        ck.out(home_dir)

    return {'return':0, 'home_dir':home_dir}


##############################################################################
# Replace string in file

def replace_string_in_file(i):
    """
    Input:  {
              file
              (file_out)  - if !='', use this file for output, otherwise overwrite original one!
              string      - string to replace
              replacement - replacement string
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              (updated)    - if 'yes', files was updated
            }

    """

    import copy

    o=i.get('out','')

    fin=i['file']
    s1=i['string']
    s2=i['replacement']

    fout=i.get('file_out','')
    if fout=='': fout=fin

    rx=ck.load_text_file({'text_file':fin})
    if rx['return']>0: return rx

    s=rx['string']
    sx=s.replace(s1,s2)

    r={'return':0, 'updated':'no'}

    if s!=sx or fin!=fout:
       r=ck.save_text_file({'text_file':fout, 'string':sx})
       r['updated']='yes'

    return r

##############################################################################
# updating json file

def refresh_json(i):
    """
    Input:  {
              json_file     - file with json
              (output_file) - if !='' use this file for output instead of rewriting original file
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    jf=i['json_file']

    of=i.get('output_file','')
    if of=='': of=jf

    r=ck.load_json_file({'json_file':jf})
    if r['return']>0: return r

    d=r['dict']

    return ck.save_json_to_file({'json_file':of, 'dict':d, 'sort_keys':'yes'})

##############################################################################
# process all files recursively using some action

def process_all_files_recursively(i):
    """
    Input:  {
               (path)     - starting path (or current)
               (pattern)  - file pattern
               (cmd)      - perform action with a file

               (ck)       - call CK access
               (file_key) - substitute this key in 'ck' with file name with full path
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    o=i.get('out','')

    p=i.get('path','')
    if p=='':
       p=os.getcwd()

    cka=i.get('ck',{})
    fk=i.get('file_key','')

    pat=i.get('pattern','')

    if o=='con':
       ck.out('')
       x=''
       if pat!='': x=' ('+pat+')'
       ck.out('Obtaining list of all files'+x+'. May take some time ...')

    r=ck.list_all_files({'path':p, 'pattern':pat, 'all':'yes'})
    if r['return']>0: return r

    lst=r['list']

    for qq in lst:
        if p=='': q=qq
        else:     q=os.path.join(p,qq)

        if len(cka)>0:
           if fk!='':
              cka[fk]=q

           if o=='con':
              ck.out('  '+q)

           r=ck.access(cka)
           if r['return']>0: return r

    return {'return':0, 'list':lst}

##############################################################################
# merge dictionaries in 2 files

def merge_dicts(i):
    """
    Input:  {
              file1   - dict1
              file2   - dict2
              (file3) - output to this file. If empty use file1
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    f1=i.get('file1','')
    f2=i.get('file2','')

    if f1=='' or f2=='':
       return {'return':1, 'error':'--file1 and --file2 should be specified'}

    fo=i.get('file3','')
    if fo=='': fo=f1

    r=ck.load_json_file({'json_file':f1})
    if r['return']>0: return r
    d1=r['dict']

    r=ck.load_json_file({'json_file':f2})
    if r['return']>0: return r
    d2=r['dict']

    if type(d1)==list and type(d2)==list:
       for q in d2:
           d1.append(q)
    else:
       r=ck.merge_dicts({'dict1':d1, 'dict2':d2})
       if r['return']>0: return r
       d1=r['dict1']

    r=ck.save_json_to_file({'json_file':fo, 'dict':d1})
    if r['return']>0: return r

    return {'return':0}

##############################################################################
# sort JSON file

def sort_json_file(i):
    """
    Input:  {
              json_file     - file with json
              (output_file) - if !='' use this file for output instead of rewriting original file
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    return refresh_json(i)

##############################################################################
# add key to meta/files of different entries

def add_key(i):
    """
    Input:  {
              data            - CID of entries to update (can be wild cards)
              (tags)          - prune entries by tags

              key             - key in flat format
              value           - value

              (ignore_update) - ignore update info in entries
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o=i.get('out','')

    data=i.get('data','')
    if data=='':
       return {'return':1, 'error':'"data" is not defined'}

    tags=i.get('tags','')

    key=i.get('key','')
    if key=='':
       return {'return':1, 'error':'"key" is not defined'}

    value=i.get('value','')
    if value=='':
       return {'return':1, 'error':'"value" is not defined'}

    iu=i.get('ignore_update','')

    # Search entries
    r=ck.access({'action':'search',
                 'cid':data,
                 'tags':tags})
    if r['return']>0: return r
    lst=r['lst']

    llst=len(lst)

    if llst>0 and o=='con':
       ck.out('Updating '+str(llst)+' entries ...')
       ck.out('')

    # Iterate over entries
    for l in lst:
        ruid=l['repo_uid']
        ruoa=l['repo_uoa']
        muid=l['module_uid']
        muoa=l['module_uoa']
        duid=l['data_uid']
        duoa=l['data_uoa']

        if o=='con':
           ck.out('* '+ruoa+':'+muoa+':'+duoa)

        # Load meta
        r=ck.access({'action':'load',
                     'repo_uoa':ruid,
                     'module_uoa':muid,
                     'data_uoa':duid})
        if r['return']>0: return r
        d=r['dict']

        # Updating dict
        r=ck.set_by_flat_key({'dict':d,
                              'key':key,
                              'value':value})
        if r['return']>0: return r

        # Store meta
        r=ck.access({'action':'update',
                     'repo_uoa':ruid,
                     'module_uoa':muid,
                     'data_uoa':duid,
                     'dict':d,
                     'substitute':'yes',
                     'ignore_update':iu})
        if r['return']>0: return r

    return {'return':0}

##############################################################################
# internal file to ignore files/directories from templates

def prepare_entry_template_ignore_files(dr, files):
    lst=['.cm','tmp']
    return lst

##############################################################################
# prepare template for a given entry
# (trying to unify templates for CK when adding program, soft, package, program, dataset, etc)

def prepare_entry_template(i):
    """
    Input:  {
              original_module_uoa - add template for this original module 
              (template)          - force using this template
              (skip_custom_note)  - if 'yes', do not print note about customization at the end

              all params from "ck add" function
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import shutil

    o=i.get('out','')
    oo=''
    if o=='con': oo=o

    omuoa=i.get('original_module_uoa','')
    ruoa=i.get('repo_uoa','')
    duoa=i.get('data_uoa','')

    # Check user-friendly name
    if duoa=='' and o=='con':
       r=ck.inp({'text': 'Enter alias for your entry without spaces such as "tool.cmake": '})
       if r['return']>0: return r
       duoa=r['string'].strip()
       i['data_uoa']=duoa
       ck.out('')

    # Search templates
    tuoa=i.get('template','')
    truoa=i.get('template_repo','')

    if tuoa=='':
       lst=[{'data_uid':'', 'data_uoa':'', 'repo_uid':'', 'info':{'data_name':'Empty entry'}, 'meta':{'sort':9999}}] # Add it to the end

       c=None

       if tuoa=='':
          if oo=='con':
             ck.out('Searching for templates ...')

          ii={'action':'search',
              'module_uoa':omuoa,
              'data_uoa':tuoa,
              'add_info':'yes',
              'add_meta':'yes',
              'search_dict':{'template':'yes'}}
          r=ck.access(ii)
          if r['return']>0: return r

          lst+=r['lst']

       if len(lst)==1:
          c=lst[0]

       # Make selection
       if oo=='con' and len(lst)>1:
          ck.out('')

          r=select_uoa({'text':'Select template for the new entry', 
                        'choices':lst})
          if r['return']>0: return r

          c=r['choice']

          ck.out('')

       if c!=None:
          tuoa=c['data_uid']
          truoa=c['repo_uid']

    d=i.get('dict',{})

    if tuoa!='':
       # Attempt to load entry
       ii={'action':'load',
           'module_uoa':omuoa,
           'data_uoa':tuoa,
           'repo_uoa':truoa}
       r=ck.access(ii)
       if r['return']>0: return r

       p=os.path.realpath(r['path'])

       d.update(r['dict'])

    i['action']='add'
    i['module_uoa']=omuoa

    i['common_func']='yes'
    i['sort_keys']='yes'

    i['dict']=d
    
    r=ck.access(i)
    if r['return']>0: return r

    pnew=r['path']

    # Copy files to a new entry if template
    if tuoa!='':
       d=os.listdir(p)
       for f in d:
           if f!='.cm' and not f.endswith('.pyc') and 'tmp' not in f:
              porig2=os.path.join(p,f)
              pnew2=os.path.join(pnew,f)

              try:

                 if os.path.isdir(porig2):
                    shutil.copytree(porig2, pnew2, ignore=shutil.ignore_patterns('.cm', '*tmp*', '*.pyc'))
                 else:
                    shutil.copyfile(porig2, pnew2)

              except IOError as e: 
                 return {'return':1, 'error':'problem copying files from template ('+str(e)+')'}

    # Print info about customization
    if i.get('skip_custom_note', '')!='yes' and oo=='con':
       ck.out('')
       ck.out('DO NOT FORGET to finish manually updating this JSON meta (tags, paths, dependencies, etc):')
       ck.out('')
       ck.out(' * JSON meta:   '+os.path.join(pnew, ck.cfg['subdir_ck_ext'], ck.cfg['file_meta']))
       ck.out(' * Other files: '+pnew)

    return r

##############################################################################
# Universal string selector.
#
# Given an ordered list of options (strings)

def select_string(i):
    """
    Input:  {
                options         - an ordered list of strings to select from
                (question)      - the question to ask
                (default)       - default selection
                (no_autoselect) - if "yes", enforce the interactive choice even if there is only 1 option
                (no_autoretry)  - if "yes", bail out on any unsuitable input, do not offer to retry
                (no_skip_line)  - if "yes", do not skip a line after each option
                (first_match)   - if "yes", take the first match in case there are multiple
            }

    Output: {
                selected_index  - an index < len(options)
                selected_value  - the string value at selected_index

                return          - return code =  0, if successful
                                              >  0, if error
                (error)         - error text if return > 0
            }

    """

    import copy

    question    = i.get('question', 'Please select from the options above')
    options     = copy.deepcopy( i.get('options') )
    default     = i.get('default', None)
    auto_select = i.get('no_autoselect', '') != 'yes'
    auto_retry  = i.get('no_autoretry', '') != 'yes'
    skip_line   = i.get('no_skip_line', '') != 'yes'
    first_match = i.get('first_match', '') == 'yes'

    if not options or len(options)==0:
        return {'return': 1, 'error': 'No options provided - please check the docstring for correct syntax'}

    num_options = len(options)

    for j in range(num_options):
        if not isinstance(options[j], list):
            options[j] = [ options[j] ]

        ck.out("{:>2}) {}".format(j, options[j][0]))
        for extra_line in options[j][1:]:
            ck.out('    {}'.format(extra_line))
        if skip_line:
            ck.out('')

    ck.out('')
    num_matches = 0

    if len(options)==1 and auto_select:
        ck.out('Since there is only one option, auto-selecting it')
        response = '0'
        selected_index = 0
        num_matches = 1

    while num_matches!=1:

        r = ck.inp({'text': "{}{}: ".format(question, ' [ hit return for "{}" ]'.format(default) if default!=None and len(default) else '')})
        response = r['string']

        if response=='' and default!=None:
            response = default

        try:                                    # try to convert into int() and see if it works
            error_message = None

            selected_index = int(response)
            if selected_index < num_options:
                num_matches = 1
            else:
                error_message = 'Selected index is out of range [0..{}]'.format(num_options-1)
        except:
            num_matches = 0
            for j in reversed(range(num_options)):
                if response in options[j][0]:
                    selected_index = j
                    num_matches += 1

            if num_matches>1 and first_match:
                break
            elif num_matches!=1:
                error_message = 'Instead of 1 unique match there were {}'.format(num_matches)

        if error_message:
            if auto_retry:
                ck.out( error_message + ", please try again" )
            else:
                return { 'return': 1, 'response': response, 'error': error_message }

    selected_value = options[selected_index][0] if selected_index >= 0 else ''

    if i.get('out')=='con':
        ck.out('You selected [{}] == "{}"'.format(selected_index, selected_value))

    return { 'return':0, 'response': response, 'selected_index': selected_index, 'selected_value': selected_value }


##############################################################################
# Search for entries of a specific type using a mix of tags AND variations
#

def search_in_variations(i):
    """
    Input:  {
                tags                - the query, a mixture of tags and variations
                (query_module_uoa)  - entries of which type to search for
                (add_info)          - each entry found will also contain 'info' dictionary ('meta' is obligatory)
            }

    Output: {
                lst             - a list of entries found (may be empty)

                return          - return code =  0, if successful
                                              >  0, if error
                (error)         - error text if return > 0
            }

    """

    tags        = i.get('tags')
    module_uoa  = i.get('query_module_uoa', '*')
    add_info    = i.get('add_info', '')

    required_tags_list = tags
    if not type(required_tags_list)==list:
        required_tags_list = required_tags_list.split(',')

    required_tags_set = set(required_tags_list)   # NB: this variable gets locked in the closure for efficiency of the following function:

    def tags_and_variations_merging_callback(i):
        """ A callback function that knows how to merge tags and variations
            and how to match the resulting set against the query.
        """

        meta = i.get('meta',{})
        supported_variations_set    = set(meta.get('variations',{}).keys())
        tags_and_variations         = set(meta.get('tags',{})) | supported_variations_set

        matched_bool                = tags_and_variations >= required_tags_set
        i['required_variations']    = list( supported_variations_set & required_tags_set )

        return { 'return': 0, 'skip': ( '' if matched_bool else 'yes' ) }

    r=ck.access({'action':'list',
                   'module_uoa': module_uoa,
                   'add_meta':'yes',
                   'add_info':add_info,
                   'download_tags':tags, # These tags are not influencing main functionality but used for auto-download of components from the CK platform
                   'filter_func_addr': tags_and_variations_merging_callback,
    })

    if r['return']>0: return r
    if i.get('out')=='con':
        for entry in r['lst']:
            print("{}:{}:{}".format(entry['repo_uoa'], entry['module_uoa'], entry['data_uoa']))

    return r


##############################################################################
# List variations of an entry or a group of entries
#

def list_variations(i):
    """
    Input:  {
                tags                - the query, a mixture of tags and variations
                (query_module_uoa)  - entries of which type to search for
                (variation_prefix)  - an optional prefix to filter the variations
                (separator)         - an optional string to separate the results (BEWARE: CK does not support trailing spaces!)
            }

    Output: {
                return          - return code =  0, if successful
                                              >  0, if error
                (error)         - error text if return > 0
            }

    """

    tags                = i.get('tags')
    query_module_uoa    = i.get('query_module_uoa', '*')
    variation_prefix    = i.get('variation_prefix', '')
    separator           = i.get('separator', '\n')

    r=ck.access({'action':'search_in_variations',
                   'module_uoa': 'misc',
                   'query_module_uoa': query_module_uoa,
                   'tags': tags,
    })
    if r['return']>0: return r

    # merge all available variations into a set:
    variations_found = set()
    for entry_dict in r['lst']:
        variations_found |= set(entry_dict['meta'].get('variations',{}).keys())

    # collect the matching ones into a list
    variations_found = [variation for variation in variations_found if variation.startswith(variation_prefix)]

    if i.get('out')=='con':
        print(separator.join(variations_found))

    return {'return':0, 'variations':variations_found}



##############################################################################
# Universal UOA selector (improved version forked 
# from ck-autotuning:module:choice and ck.kernel)

def select_uoa(i):
    """
    Input:  {
              choices      - list from search function
              (text)       - selection text
              (skip_enter) - if 'yes', do not select 0 when user presses Enter
              (skip_sort)  - if 'yes', do not sort array
            }

    Output: {
              return  - return code =  0, if successful
                                    >  0, if error
              (error) - error text if return > 0
              choice  - {dict of selection from lst}
            }

    """

    se=i.get('skip_enter','')

    lst=i.get('choices',[])

    # Prepare data_name and then data_uoa

    if i.get('skip_sort','')!='yes':
       slst=sorted(lst, key=lambda v: (v.get('meta',{}).get('sort',0), v.get('info',{}).get('data_name',''), v['data_uoa']))
    else:
       slst=lst

    array={}
    n=0

    for x in slst:
        sn=str(n)
        array[sn]=x

        duoa=x['data_uoa']
        name=x.get('info',{}).get('data_name','')
        if name=='': name=duoa

        s=sn+') '+name
        if duoa!='': s+=' (--template='+duoa+')'

        ck.out(s)

        n+=1

    ck.out('')

    text=i.get('text','')
    if text=='': text='Select UOA'

    s=text
    if se!='yes': s+=' (or press Enter for 0)'
    s+=': '

    rx=ck.inp({'text':s})
    y=rx['string'].strip()

    if y=='' and se!='yes': y='0' 

    if y not in array:
       return {'return':1, 'error':'number is not recognized'}

    return {'return':0, 'choice':array[y]}

##############################################################################
# list CK kernel functions

def list_kernel_functions(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import copy

    o=i.get('out','')

    of=i.get('out_file','')
    if of!='':
       xof=os.path.splitext(of)

    html=False
    if o=='html' or i.get('web','')=='yes':
       html=True

    h=''
    h2=''
    hcfg=''
    if i.get('new','')=='yes':
       ii=copy.deepcopy(i)
       ii['ck_title']='Shared CK kernel functions and configurations'
       r=preload_html_for_lists(ii)
       if r['return']>0: return r

       h=r['html_start']+'\n'
       h2=r['html_stop']+'\n'

       hcfg=h

    p=ck.work['env_root'] # Internal CK path

    pk=os.path.join(p, 'ck', 'kernel.py')
    if not os.path.isfile(pk):
       return {'return':1, 'error':'Can\'t find kernel in '+pk}

    r=ck.load_text_file({'text_file':pk, 'split_to_list':'yes'})
    if r['return']>0: return r

    l=r['lst']

    funcs={}
    desc=[]
    ld=0
    target=''

    cfg=[]
    cfg_start=False

    ll=len(l)
    for k in range(0, ll):
        x=l[k]

        if x.startswith('cfg={'):
           cfg_start=True

        if cfg_start:
           cfg.append(x)
           if x=='    }':
              cfg_start=False

        if x.startswith('##################'):
           desc=[]
           ld=k

        if x.startswith('# '):
           x1=x[2:].strip()
           if x1.startswith('TARGET: '):
              target=x1[8:]
           else:
              desc.append(x1)

        if x.startswith('def '):
           j1=x.find('(')
           j2=x.find(')', j1+1)
           j3=x.find(':')

           fn=x[4:j1]
           i=x[j1+1:j2]
           rem=x[j3+1:]

           api=[]

           k+=1

           found=False
           first=False
           while not found or k<ll:
              x=l[k]
              if x.strip().startswith('"""'):
                 if first:
                    found=True
                    break
                 else:
                    first=True
              elif first:
                 api.append(x)
              k+=1
           
           if found:
              funcs[fn]={'api':api,
                         'input':i,
                         'rem':rem,
                         'desc':desc,
                         'line':ld,
                         'target':target}

           api=[]
           desc=[]
           target=''

    # Process functions
    if html:
       h+='We encourage you to reuse these portable productivity functions (Python 2.6+ and 3+) from the CK modules or in your own Python scripts:\n'
       h+='\n'
       h+='<pre>\n'
       h+='    import ck.kernel as ck\n'
       h+='\n'
       h+='    ck.out("Hello world")\n'
       h+='\n'
       h+='    r=ck.access({"action":"ls", "module_uoa":"env", "out":"con"})\n'
       h+='    if r["return"]>0: ck.err(r)\n'
       h+='    print (r["lst"])\n'
       h+='\n'
       h+='    r=ck.load_json_file({"json_file":"my_file.json", "dict":{"test":"yes"}})\n'
       h+='    if r["return"]>0: return r\n'
       h+='\n'
       h+='    r=ck.save_text_file({"text_file":"my_file.txt", "string":"test=yes\\n"})\n'
       h+='    if r["return"]>0: return r\n'
       h+='\n'
       h+='</pre>\n'

       h+='See <a href="https://github.com/ctuning/ck/wiki">CK documentation for further details</a>.\n'

       h+='<p>\n'
       h+='<table cellpadding="4" border="1" style="border-collapse: collapse; border: 1px solid black;">\n'

       h+=' <tr>\n'
       h+='  <td nowrap><b>#</b></td>\n'
       h+='  <td nowrap><b>Function name</b></td>\n'
       h+='  <td nowrap><b>Note and API</b></td>\n'
       h+=' </tr>\n'

    hdev=h

    num1=0
    num2=0
    for f in sorted(funcs):
        x=funcs[f]

        api=x['api']
        i=x['input']
        rem=x['rem']
        line=x['line']
        desc=x['desc']
        target=x['target']

        xapi='<i>'
        for y in desc:
            j=y.find('\\n=')
            if j>=0: 
               y=y[:j]
            xapi+=y
        xapi+='</i>\n'

        xapi+='<p>\n'
        xapi+='<pre>\n'
        for y in api:
            xapi+=y+'\n'
        xapi+='</pre>\n'

        url='https://github.com/ctuning/ck/blob/master/ck/kernel.py'

        x=''
        if target!='':
           x+='<p>&nbsp;&nbsp;&nbsp;<i>for '+target+'</i>'

        if 'end users' in target or 'end-users' in target:
           num1+=1
           num=num1
        else:
           num2+=1
           num=num2

        zh=' <tr>\n'
        zh+='  <td nowrap valign="top"><a name="'+f+'">'+str(num)+'</td>\n'
        zh+='  <td nowrap valign="top"><a href="'+url+'#L'+str(line+1)+'"><b>ck.'+f+'('+i+')</b></a>'+x+'</b></td>\n'
        zh+='  <td nowrap valign="top">'+xapi+'</td>\n'
        zh+=' </tr>\n'

        if 'end users' in target or 'end-users' in target:
           h+=zh
        else:
           hdev+=zh

    # Prepare config
    hcfg+='You can access the following CK internal variables:\n'
    hcfg+='\n'
    hcfg+='<pre>\n'
    hcfg+='    import ck.kernel as ck\n'
    hcfg+='\n'
    hcfg+='    print (ck.cfg)\n'
    hcfg+='</pre>\n'

    hcfg+='See <a href="https://github.com/ctuning/ck/wiki">CK documentation for further details</a>.\n'

    hcfg+='<p>\n'
    hcfg+='<pre>\n'
    for x in cfg:
        hcfg+=x+'\n'
    hcfg+='</pre>\n'
    hcfg+=h2

    if html:
       h+='</table>\n'
       h+=h2
       hdev+='</table>\n'
       hdev+=h2

       if of!='':
          r=ck.save_text_file({'text_file':of+'.html', 'string':h})
          if r['return']>0: return r

          r=ck.save_text_file({'text_file':of+'-dev.html', 'string':hdev})
          if r['return']>0: return r

          r=ck.save_text_file({'text_file':of+'-dev-cfg.html', 'string':hcfg})
          if r['return']>0: return r

    return {'return':0}

##############################################################################
# list repositories

def list_repos(i):
    """
    Input:  {
               (the same as ck search; can use wildcards)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import copy

    o=i.get('out','')

    of=i.get('out_file','')
    if of!='':
       xof=os.path.splitext(of)

    html=False
    if o=='html' or i.get('web','')=='yes':
       html=True

    h=''
    h2=''
    if i.get('new','')=='yes':
       ii=copy.deepcopy(i)
       ii['ck_title']='Shared CK repositories'
       r=preload_html_for_lists(ii)
       if r['return']>0: return r

       h=r['html_start']+'\n'
       h2=r['html_stop']+'\n'

    unique_repo=False
    if i.get('repo_uoa','')!='': unique_repo=True

    import copy
    ii=copy.deepcopy(i)

    ii['out']=''
    ii['module_uoa']=cfg['module_deps']['repo']
    ii['action']='list'
    ii['add_meta']='yes'
    ii['time_out']=-1

    rx=ck.access(ii)
    if rx['return']>0: return rx

    ll=rx['lst']

    if html:
       h+='<h2>Please check our new <a href="http://ReuseResearch.com/c.php?c=repo">beta browser</a> for CK components!</h2>\n'

       h+='<br>\n'

       h+='<p>Users share their reusable CK components (code, data, models) and workflows/pipelines using CK-compatible repositories and archives\n'
       h+='(for example, see this <a href="https://github.com/dividiti/ck-rpi-optimization">CK repository</a>\n'
       h+='with a complete compiler crowd-tuning workflow, an <a href="http://cKnowledge.org/rpi-crowd-tuning">automatically generated interactive article</a>,\n'
       h+='and a <a href="http://cKnowledge.org/dashboard">public interactive dashboard</a>).\n'
       h+='Most of them are shared with the permissive 3-clause BSD license or CC-BY to accelerate technology transfer!\n'

       h+='<p>You can run and reuse workflows, <a href="http://cKnowledge.org/shared-modules.html">modules</a> and other components with a common API from below repositories:\n'
       h+='<pre>\n'
       h+=' ck pull repo:{Repo UOA - see below}\n'
       h+='</pre>\n'

       h+='<p>You can add dependency on a given repository in your own CK repository by editing your .ckr.json file as follows:\n'
       h+='<pre>\n'
       h+=' {\n'
       h+='   ...\n'
       h+='   "dict": {\n'
       h+='      "repo_deps": [\n'
       h+='         {\n'
       h+='           "repo_uoa": "ck-tensorflow",\n'
       h+='           "repo_url": "https://github.com/ctuning/ck-tensorflow"\n'
       h+='         }\n'
       h+='         ...\n'
       h+='      ]\n'
       h+='      ...\n'
       h+='   }\n'
       h+='}\n'
       h+='</pre>\n'

       h+='<p>Feel free to add description of your own CK repository to <a href="https://github.com/ctuning/ck-env/blob/master/cfg/list-of-repos/.cm/meta.json">this JSON file</a> - just open a PR on this <a href="https://github.com/ctuning/ck-env">GitHub repo</a>.\n'
       h+='You can obtain the main key for your entry (DATA UID) as well as data_uoa (DATA ALIAS) using this command: ck info repo:{my repo name}. You can then obtain and add "dict" using this command: ck load repo:{my repo name} --min .\n'

       h+='<p>See <a href="https://github.com/ctuning/ck/wiki">CK documentation</a>,\n'
       h+=' <a href="https://github.com/ctuning/ck/wiki#contributing">"how to contribute" guide</a>\n'
       h+=' and <a href="https://github.com/ctuning/ck/wiki#user-content-reusable-ck-components">already shared reusable components</a> for further details.\n'

       h+='<p>\n'
       h+='<table cellpadding="4" border="1" style="border-collapse: collapse; border: 1px solid black;">\n'

       h+=' <tr>\n'
       h+='  <td nowrap><b>#</b></td>\n'
       h+='  <td nowrap><b>Repository UOA</b></td>\n'
       h+='  <td nowrap><b>Workflow</b></td>\n'
       h+='  <td><b>Description</b></td>\n'
       h+=' </tr>\n'

    repos={}
    repo_url={}
    repo_private={}

    # Check which already manually added from cfg:list-of-repos
    rruoa=i.get('repo_uoa','')

    r=ck.access({'action':'load',
                 'module_uoa':cfg['module_deps']['cfg'],
                 'data_uoa':cfg['cfg-list-of-repos']})
    if r['return']>0 and r['return']!=16: return r
    if r['return']==0:
       repos=r['dict']
       if rruoa=='': rruoa=r['repo_uid']

    # Going through already described repos and add missing ones
    private=''
    remote=''
    for l in ll:

        lr=l['data_uoa']
        lr_uid=l['data_uid']

        url=''

        if lr=='default':
           url='https://github.com/ctuning/ck/tree/master/ck/repo'
        else:
           rx=ck.load_repo_info_from_cache({'repo_uoa':lr_uid})
           if rx['return']>0: return rx
           url=rx.get('dict',{}).get('url','')
           private=rx.get('dict',{}).get('private','')
           remote=rx.get('dict',{}).get('remote','')

        if lr not in cfg.get('skip_repos',[]) and remote!='yes' and private!='yes' and url!='':
           if lr_uid not in repos:
              repos[lr_uid]={'data_uoa':lr}

    # Try to fill in missing information or find duplicates
    to_delete=[]
    for repo_uid in list(repos.keys()):
        x=repos[repo_uid]

        repo_uoa=x['data_uoa']

        d=x.get('dict',{})
      
        # Find real repo and get .ckr.json
        rx=ck.access({'action':'where',
                      'module_uoa':cfg['module_deps']['repo'],
                      'data_uoa':repo_uoa})
        if rx['return']==0: 
           pckr=os.path.join(rx['path'], ck.cfg['repo_file'])
           if os.path.isfile(pckr):
              rx=ck.load_json_file({'json_file':pckr})
              if rx['return']>0: return rx

              rxd=rx['dict']

              dx=rxd['dict']

              if 'path' in dx:
                 del(dx['path'])

              real_repo_uid=rxd['data_uid']

              # Check if mismatch of real uid and current one (old bug - should be fixed now)
              if real_repo_uid!=repo_uid:
                 ck.out('')
                 ck.out('WARNING: repo UID mismatch for '+repo_uoa+' ('+real_repo_uid+' != '+repo_uid+')')
                 ck.out('')

#                 r=ck.inp({'text':'  update correct one and remove old (Y/n)?'})
#                 if r['return']>0: return r
#                 xx=r['string'].strip().lower()
                 if True: #xx!='n':
                    if real_repo_uid not in repos:
                       repos[real_repo_uid]=copy.deepcopy(repos[repo_uid])
                    else:
                       if not 'dict' in repos[real_repo_uid]:
                          repos[real_repo_uid]['dict']={}
                       repos[real_repo_uid]['dict'].update(d)

                    to_delete.append(repo_uid)
              else:
                 # light update
                 for k in dx:
                     if k not in d:
                        d[k]=dx[k]

    # Delete duplicates (old bug)
    for tod in to_delete:
        if tod in repos: 
           ck.out('* Deleting '+tod)
           del(repos[tod])

    # Record list
    if rruoa=='': rruoa='ck-env'

    r=ck.access({'action':'update',
                 'module_uoa':cfg['module_deps']['cfg'],
                 'data_uoa':cfg['cfg-list-of-repos'],
                 'repo_uoa':rruoa,
                 'dict':repos,
                 'substitute':'yes',
                 'sort_keys':'yes',
                 'ignore_update':'yes'})
    if r['return']>0: return r
    
    # Show repos
    num=0
    for repo in sorted(repos, key=lambda k: repos[k]['data_uoa']):
        l=repos[repo]

        lr=l['data_uoa']
        lr_uid=repo

        d=l.get('dict',{})

        if d.get('skip_from_index','')=='yes':
           continue

        url=d.get('url','')
        external_url=d.get('external_url','')
        rd=d.get('repo_deps',{})

        ld=d.get('desc','')

        workflow_desc=d.get('workflow_desc','')

        to_get=''
        if url.find('github.com/ctuning/')>0:
           to_get='ck pull repo:'+lr
        elif url!='':
           to_get='ck pull repo --url='+url
        elif external_url!='':
           to_get='[ <a href=\"'+external_url+'\">external link</a> ]'

        num+=1

        ###############################################################
        if html:
           h+=' <tr>\n'

           x1=''
           x1readme=''
           x2=''
           z1=''
           z11=''
           url2=''

           if url!='' and url.startswith('git@'):
              url=url.replace(':','/').replace('git@','https://')

           if url=='' and lr=='default':
              url='https://github.com/ctuning/ck/tree/master/ck/repo'

           if url!='':
              x1='<a href="'+url+'">'
              x1readme='<a href="'+url+'#readme">'
              x2='</a>'

              url2=url

              if url2.endswith('.git'):
                 url2=url2[:-4]

              if '/tree/master/' not in url2:
                 url2+='/tree/master/'
              else:
                 url2+='/'

              z1='<a href="'+url2+'.ckr.json">'

           # TD
           h+='  <td nowrap valign="top"><a name="'+lr+'">'+str(num)+'</b></td>\n'
                                                     
           # TD
           x5=''
           if url!='':
              x5='<br><ul>( '+x1readme+'README'+x2+' ; '+z1+'JSON meta</a> )'

           h+='  <td nowrap valign="top">'+x1+lr+x2+''+x5+'</td>\n'

           # TD
           workflow_desc=workflow_desc.replace('$#repo_url#$',url2)

           if d.get('ck_artifact','')!='' or d.get('reproducible_article','')=='yes' or d.get('passed_artifact_evaluation','')=='yes':
              if workflow_desc!='': workflow_desc+='<p>'
              workflow_desc+='reproducible&nbsp;paper\n'
              if d.get('passed_artifact_evaluation','')=='yes':
                 workflow_desc+='-&nbsp;passed&nbsp;<a href="http://cTuning.org/ae">Artifact&nbsp;Evaluation</a>:\n'
                 workflow_desc+='<p><center><img src="https://www.acm.org/binaries/content/gallery/acm/publications/replication-badges/artifacts_evaluated_reusable_dl.jpg" width="64"></center>\n'

           h+='  <td valign="top"><center>\n'
           h+='    '+workflow_desc+'\n'
           h+='  </center></td>\n'

           # TD
           ld=ld.replace('$#repo_url#$',url2)
           h+='  <td valign="top">'+ld+'\n'

           if to_get!='':
              h+='<p>\n'
              h+='How to get:\n'
              h+='<pre>\n'
              h+=to_get+'\n'
              h+='</pre>\n'

           if len(rd)>0:
              h+='<p>Dependencies on other repositories:\n'
              h+='<ul>\n'
              for qq in sorted(rd, key=lambda k: k['repo_uoa']):
                  repo_uoa=qq['repo_uoa']
                  repo_url=qq.get('repo_url','')

                  if repo_url=='':
                     repo_url='https://github.com/ctuning/'+repo_uoa

                  if repo_url!='':
                     repo_uoa='<a href="#'+repo_uoa+'">'+repo_uoa+'</a>'

                  h+='<li>'+repo_uoa+'\n'
              h+='</ul>\n'

           h+='</td>\n'

           h+=' </tr>\n'

    ck.out('')
    ck.out('  Total repos: '+str(num))
    ck.out('')

    if html:
       h+='</table>\n'
       h+=h2

       if of!='':
          r=ck.save_text_file({'text_file':of, 'string':h})
          if r['return']>0: return r

    return {'return':0}

##############################################################################
# list modules

def list_modules(i):
    """
    Input:  {
              (new) - if 'yes', add htmls
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import copy

    o=i.get('out','')

    of=i.get('out_file','')
    if of!='':
       xof=os.path.splitext(of)

    html=False
    if o=='html' or i.get('web','')=='yes':
       html=True

    h=''
    h2=''
    if i.get('new','')=='yes':
       ii=copy.deepcopy(i)
       ii['ck_title']='Shared CK modules'
       r=preload_html_for_lists(ii)
       if r['return']>0: return r

       h=r['html_start']+'\n'
       h2=r['html_stop']+'\n'

    unique_repo=False
    if i.get('repo_uoa','')!='': unique_repo=True

    ii=copy.deepcopy(i)

    ii['out']=''
    ii['action']='list'
    ii['module_uoa']=cfg['module_deps']['module']
    ii['add_meta']='yes'
    ii['time_out']=-1

    rx=ck.access(ii)
    if rx['return']>0: return rx

    ll=sorted(rx['lst'], key=lambda k: k['data_uoa'])

    if html:
       h+='<h2>Please check our new <a href="http://ReuseResearch.com/c.php?c=program">beta browser</a> for CK components!</h2>\n'

       h+='<br>\n'

       h+='You can install and reuse CK modules as follows:\n'
       h+='<pre>\n'
       h+=' ck pull repo:{Repo UOA - see below}\n'
       h+=' ck help {module UOA - see below}\n'
       h+='</pre>\n'

       h+='You can check a JSON API of a given action of a given module as follows:\n'
       h+='<pre>\n'
       h+=' ck {module action - see below} {module UOA} --help\n'
       h+='</pre>\n'

       h+='You can add your own dummy CK module as follows:\n'
       h+='<pre>\n'
       h+=' ck add module:{my module alias}\n'
       h+='</pre>\n'

       h+='You can add a new action to the CK module as follows:\n'
       h+='<pre>\n'
       h+=' ck add_action module:{my module alias}\n'
       h+='</pre>\n'

       h+='See <a href="https://github.com/ctuning/ck/wiki">CK documentation</a>,\n'
       h+=' <a href="https://github.com/ctuning/ck/wiki#contributing">"how to contribute" guide</a>,\n'
       h+=' and the latest <a href="http://cKnowledge.org/rpi-crowd-tuning">CK paper</a> for further details.\n'

       h+='<p>\n'
       h+='<table cellpadding="4" border="1" style="border-collapse: collapse; border: 1px solid black;">\n'

       h+=' <tr>\n'
       h+='  <td nowrap><b>#</b></td>\n'
       h+='  <td nowrap><b>Module&nbsp;UOA with JSON API<br>(Python module/wrapper/plugin)</b></td>\n'
       h+='  <td nowrap><b>Repo UOA</b></td>\n'
       h+='  <td nowrap><b>Workflow (pipeline)?</b></td>\n'
       h+='  <td><b>Description and actions</b></td>\n'
       h+=' </tr>\n'

    repo_url={}
    repo_private={}

    private=''
    num=0

    for l in ll:
        ln=l['data_uoa']
        lr=l['repo_uoa']

        lr_uid=l['repo_uid']
        url=''
        if lr=='default':
           url='https://github.com/ctuning/ck/tree/master/ck/repo'
        elif lr_uid in repo_url:
           url=repo_url[lr_uid]
        else:
           rx=ck.load_repo_info_from_cache({'repo_uoa':lr_uid})
           if rx['return']>0: return rx
           url=rx.get('dict',{}).get('url','')
           repo_private[lr_uid]=rx.get('dict',{}).get('private','')
           repo_url[lr_uid]=url

        private=repo_private.get(lr_uid,'')

#        if lr not in cfg.get('skip_repos',[]) and private!='yes' and url!='':
        if lr not in cfg.get('skip_repos',[]) and private!='yes' and url!='':
           num+=1

           ck.out('  '+str(num)+') '+ln)

           lm=l['meta']
           ld=lm.get('desc','')

           xworkflow=lm.get('workflow','')
           workflow=lm.get('workflow_type','')
           if xworkflow=='yes' and workflow=='':
              workflow='yes'

           actions=lm.get('actions',{})

           if lr=='default':
              to_get=''
           elif url.find('github.com/ctuning/')>0:
              to_get='ck pull repo:'+lr
           else:
              to_get='ck pull repo --url='+url

           ###############################################################
           if html:
              h+=' <tr>\n'

              x1=''
              x2=''
              z1=''
              z1full=''
              z11=''
              if url!='':
                 x1='<a href="'+url+'">'
                 x2='</a>'

                 url2=url

                 if url2.endswith('.git'):
                    url2=url2[:-4]

                 if '/tree/master/' not in url2:
                    url2+='/tree/master/module/'
                 else:
                    url2+='/module/'

                 z1full=url2+ln+'/module.py'
                 z1='<a href="'+z1full+'">'
                 z11='<a href="'+url2+ln+'/.cm/meta.json">'

              h+='  <td nowrap valign="top"><a name="'+ln+'">'+str(num)+'</b></td>\n'

              h+='  <td nowrap valign="top">'+z1+ln+x2+'</b> <i>('+z11+'CK meta'+x2+')</i></td>\n'

              h+='  <td nowrap valign="top"><b>'+x1+lr+x2+'</b></td>\n'

              h+='  <td nowrap valign="top">'+workflow+'</td>\n'

              h+='  <td valign="top">'+ld+'\n'

              if len(actions)>0:
                 h+='<ul>\n'
                 for q in sorted(actions):

                     qq=actions[q]
                     qd=qq.get('desc','')

                     h+='<li>"ck <i>'+q+'</i> '+ln+'"'
                     if qd!='':
                        h+=' - '+qd

                     if z1full!='':
                        # Get API!
                        l=-1
                        rx=ck.get_api({'module_uoa':ln, 'func':q})
                        if rx['return']==0:
                           l=rx['line']

                        if l!=-1:
                           h+=' (&nbsp;<a href="'+z1full+'#L'+str(l)+'">API</a>&nbsp;)'
                 h+='</ul>\n'

              h+='</td>\n'

              h+=' </tr>\n'

           ###############################################################
           elif o=='mediawiki':
              x=lr
              if url!='':
                 x='['+url+' '+lr+']'
              ck.out('')
              ck.out('=== '+ln+' ('+lr+') ===')
              ck.out('')
              ck.out('Desc: '+ld)
              ck.out('<br>CK Repo URL: '+x)
              if to_get!='':
                 ck.out('<br>How to get: <i>'+to_get+'</i>')
              ck.out('')
              if len(actions)>0:

                 ck.out('Actions (functions):')
                 ck.out('')

                 for q in sorted(actions):
                     qq=actions[q]
                     qd=qq.get('desc','')
                     ck.out('* \'\''+q+'\'\' - '+qd)

           ###############################################################
           elif o=='con' or o=='txt':
              if unique_repo:
                 ck.out('')
                 s=ln+' - '+ld

              else:
                 ss=''
                 if len(ln)<35: ss=' '*(35-len(ln))

                 ss1=''
                 if len(lr)<30: ss1=' '*(30-len(lr))

                 s=ln+ss+'  ('+lr+')'
                 if ld!='': s+=ss1+'  '+ld

              ck.out(s)

              if len(actions)>0:
                 ck.out('')
                 for q in sorted(actions):
                     qq=actions[q]
                     qd=qq.get('desc','')
                     ck.out('  * '+q+' - '+qd)


    ck.out('')
    ck.out('  Total modules: '+str(num))
    ck.out('')

    if html:
       h+='</table>\n'
       h+=h2

       if of!='':
          r=ck.save_text_file({'text_file':of, 'string':h})
          if r['return']>0: return r

    return {'return':0}

##############################################################################
# preload HTMLs for lists of components

def preload_html_for_lists(i):
    """
    Input:  {
              (html_file_start) - ck_start_html by default
              (html_file_stop)  - ck_stop_html by default
              (ck_title)        - update title in the start file
              (out_file)        - get page name
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html_start
              html_stop
            }

    """

    import os

    out_file=i.get('out_file','')

    page_name=os.path.basename(out_file)
    if page_name.endswith('.html'):
       page_name=page_name[:-5]

    fstart=i.get('html_file_start','')
    if fstart=='': fstart='ck_start.html'

    fstop=i.get('html_file_stop','')
    if fstop=='': fstop='ck_stop.html'

    ck_title=i.get('ck_title','')

    # Load first file
    r=ck.load_text_file({'text_file':fstart})
    if r['return']>0: return r

    html_start=r['string'].replace('$#ck_title#$',ck_title)
    html_start=html_start.replace('$#ck_page#$',page_name)

    # Load second file
    r=ck.load_text_file({'text_file':fstop})
    if r['return']>0: return r

    html_stop=r['string'].replace('$#ck_title#$',ck_title)

    return {'return':0, 'html_start':html_start, 'html_stop':html_stop}

##############################################################################
# replace multiple strings in a given file

def replace_strings_in_file(i):
    """
    Input:  {
              file
              (file_out)            - if !='', use this file for output, otherwise overwrite original one!
              replacement_json_file - replacement file with multiple strings to substitute
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              (updated)    - if 'yes', files was updated
            }

    """

    import copy

    o=i.get('out','')

    fin=i['file']
    rjf=i['replacement_json_file']

    fout=i.get('file_out','')
    if fout=='': fout=fin

    rx=ck.load_text_file({'text_file':fin})
    if rx['return']>0: return rx
    s=rx['string']

    rx=ck.load_json_file({'json_file':rjf})
    if rx['return']>0: return rx
    rep=rx['dict']

    sx=s
    for k in rep:
        v=rep[k]
        sx=sx.replace(k,v)

    r={'return':0, 'updated':'no'}

    if s!=sx or fin!=fout:
       r=ck.save_text_file({'text_file':fout, 'string':sx})
       r['updated']='yes'

    return r


def transfer(i):
    """
    Input:  {
                xcids[]             - (remote) entries to be copied to a local repository, made of repo_uoa:module_uoa:data_uoa

                (source_repo_uoa)   - in case the source is remote, the specific repository on the source (optional - only needed for disambiguation)
                (target_server_uoa) - the optional target remote server (mapped as a local repository)
                (target_repo_uoa)   - the repository to store the entries on target server ('local' by default)
                (update_meta_dict)  - the dictionary with which to update the original meta.json
                (update_mmeta_dict) - the dictionary with which to update the meta dictionary inside meta.json
                (tags)              - filter the source list by tags
            }

    Output: {
                return      - return code =  0, if successful
                                          >  0, if error
                (error)     - error text if return > 0
            }
    """

    import os

    o                   = i.get('out','')
    source_addrs        = i.get('xcids', [])       # can unshift the "default entry" into this list when/if moving this method to the kernel

    source_repo_uoa     = i.get('source_repo_uoa')
    target_server_uoa   = i.get('target_server_uoa')
    target_repo_uoa     = i.get('target_repo_uoa', 'local')
    update_meta_dict    = i.get('update_meta_dict', {})
    update_mmeta_dict    = i.get('update_mmeta_dict', {})

    tags                = i.get('tags')

    if len(source_addrs)==0:
        return {'return':1, 'error': 'Need a non-empty list of source CID addresses'}

    expanded_source_addrs = {}

    for addr_pattern in source_addrs:
        search_adict = {'action':       'search',
        }
        search_adict.update( addr_pattern )
        if tags:
            search_adict['tags']            = tags
        if source_repo_uoa:
            search_adict['remote_repo_uoa'] = source_repo_uoa

        r=ck.access( search_adict )
        if r['return']>0: return r

        for found_addr in r['lst']:     # NB: remote found_addr comes with screwed up repo_uoa, we have to undo that
            source_addr = {
                'data_uoa':     found_addr['data_uoa'],
                'module_uoa':   found_addr['module_uoa'],
            }
            if addr_pattern.get('repo_uoa') and found_addr.get('repo_uoa')!=addr_pattern.get('repo_uoa'):
                source_addr.update( {
                    'remote_repo_uoa':    found_addr.get('repo_uoa'),
                    'repo_uoa':         addr_pattern.get('repo_uoa'),
                })
            else:
                source_addr['repo_uoa'] = found_addr.get('repo_uoa')

            cid = ':'.join([ source_addr['repo_uoa'], source_addr['module_uoa'], source_addr['data_uoa'] ])
            if cid in expanded_source_addrs:
                return {'return':2, 'error': "Entry {} was found on the source multiple times - please use --source_repo_uoa to disambiguate!".format(cid)}
            else:
                expanded_source_addrs[cid] = source_addr

    for source_addr in expanded_source_addrs.values():
        from_local = source_addr.get('remote_repo_uoa') == None
        if from_local and (not target_server_uoa):
            return {'return':3, 'error': 'Cannot copy the entries locally (since IDs are to be preserved)'}

        target_addr = {
            'module_uoa':       source_addr['module_uoa'],
            'data_uoa':         source_addr['data_uoa'],
        }
        if target_server_uoa:
            target_addr.update({
                'remote_repo_uoa':  target_repo_uoa,
                'repo_uoa':         target_server_uoa,
            })
        else:
            target_addr['repo_uoa'] = target_repo_uoa


        load_adict = {  'action':           'load',
        }
        load_adict.update( source_addr )
        r=ck.access( load_adict )
        if r['return']>0: return r

        meta_dict           = r['dict']
        data_uid            = r['data_uid']

        meta_dict.update( update_meta_dict )
        meta_dict.get('meta', {}).update( update_mmeta_dict )

        add_adict = {   'action':           'add',
                        'common_func':      'yes',
                        'dict':             meta_dict,              # copying meta data
                        'data_uid':         data_uid,               # copying the original data_uid
        }
        add_adict.update( target_addr )
        r=ck.access( add_adict )
        if r['return']>0: return r

        pull_adict = {  'action':           'pull',
                        'archive':          'yes',
        }
        pull_adict.update( source_addr )
        r=ck.access( pull_adict )
        if r['return']>0: return r

        zip_name = ck.cfg['default_archive_name']

        push_adict = {  'action':           'push',
                        'archive':          'yes',
                        'filename':         zip_name,
        }
        push_adict.update( target_addr )
        r=ck.access( push_adict )
        if r['return']>0: return r

        if os.path.isfile(zip_name):
            os.remove(zip_name)     # to avoid clashing with the next one

        if o=='con':
            display_source_repo = source_addr['repo_uoa'] if from_local else '{}/{}'.format(source_addr['repo_uoa'], source_addr['remote_repo_uoa'])
            display_target_repo = '{}/{}'.format(target_server_uoa, target_repo_uoa) if target_server_uoa else target_repo_uoa
            display_mod_data    = '{}:{}'.format(source_addr['module_uoa'], source_addr['data_uoa'])
            ck.out('{}:{} -> {}:{}'.format(display_source_repo, display_mod_data, display_target_repo, display_mod_data))

    return {'return':0}


def clone_server_repo(i):   # FIXME: it's probably better to zip the whole thing and transfer in one go
    """
    Input:  {
                source_repo_uoa     - which remote repo to clone

                (target_repo_uoa)   - the optional local name for the new repo (defaults to source_repo_uoa)
            }

    Output: {
                return      - return code =  0, if successful
                                          >  0, if error
                (error)     - error text if return > 0
            }
    """

    o                   = i.get('out','')
    source_repo_uoa     = i.get('source_repo_uoa')

    if not source_repo_uoa:
        return {'return':1, 'error': 'source_repo_uoa is the obligatory parameter'}

    target_repo_uoa     = i.get('target_repo_uoa', source_repo_uoa)


    add_repo_adict = {  'action':       'add',
                        'module_uoa':   'repo',
                        'data_uoa':     target_repo_uoa,
                        'quiet':        'yes',
                        'out':          o,
    }

    r=ck.access( add_repo_adict )
    if r['return']>0: return r

    transfer_adict = {  'action':           'transfer',
                        'module_uoa':       'misc',
                        'cids':             [ 'remote-ck:*:*' ],    # sic: you call with unparsed 'cids' and the method will see parsed 'xcids'
                        'source_repo_uoa':  source_repo_uoa,
                        'target_repo_uoa':  target_repo_uoa,
                        'out':              o,
    }
    r=ck.access( transfer_adict )
    if r['return']>0: return r

    return {'return':0}


def substitute_from_dict(i):
    """
    Input:  {
                input_string        - original string with placeholders for substitution
                mapping             - dictionary to subsitute from
                (input_open)        - opening marker for the keyword in the input_string
                (input_close)       - closing marker for the keyword in the input_string
                (tolerant)          - if "yes", do not fail if a keyword is missing from mapping
            }

    Output: {
                return      - return code =  0, if successful
                                          >  0, if error
                (error)     - error text if return > 0
            }
    Tests:
        ck substitute_from_dict misc --input_string='alpha $<<FOO>>$ gamma' "@@@{'mapping': {'FOO': 'beta'}}"

        ck substitute_from_dict misc --input_string='alpha$#dir_sep#$gamma $#foo#$X' \
            "@@@{'mapping': {'dir_sep': '/'}}" --input_open='$#' --input_close='#$' --tolerant
    """

    import re

    input_string    = i['input_string']
    mapping         = i['mapping']

    input_open      = i.get('input_open', '$<<')
    input_close     = i.get('input_close', '>>$')
    tolerant        = i.get('tolerant') == 'yes'

    pattern         = re.compile( '({}(\w+){})'.format(re.escape(input_open), re.escape(input_close)) )
    output_string   = input_string

    for match in re.finditer(pattern, input_string):
        expression, keyword = match.group(1), match.group(2)
        if keyword in mapping:
            output_string = output_string.replace(expression, mapping[keyword])
        elif tolerant:
            output_string = output_string.replace(expression, "")
        else:
            return {'return':1, 'error': "Could not map {} from the given dictionary".format(expression) }

    if i.get('out') == 'con':
        ck.out(output_string)

    return {'return': 0, 'output_string': output_string}


def capture_command_output(i):
    """
    Input:  {
                cmd                 - shell command with parameters to run
            }

    Output: {
                output_lines        - an array of lines of the command's output

                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    Tests:
        ck capture_command_output misc --cmd='ls -1'
    """

    import subprocess

    cmd = i['cmd']

    if type(cmd)!=list:
        cmd = cmd.split()   # split on general whitespace

    try:
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE)

        stdout_lines = out.communicate()[0].decode("utf8").split("\n")

        if stdout_lines[-1] == '':
            stdout_lines.pop()  # remove the last empty element

        return {'return': 0, 'output_lines': stdout_lines}

    except:
        return {'return': 1, 'error': "Could not run the command '{}'".format(cmd) }

##############################################################################
# list CK kernel functions (md format)

def list_kernel_functions_md(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import copy

    o=i.get('out','')

    of=i.get('out_file','')

    h=''
    h2=''
    hcfg=''

    p=ck.work['env_root'] # Internal CK path

    pk=os.path.join(p, 'ck', 'kernel.py')
    if not os.path.isfile(pk):
       return {'return':1, 'error':'Can\'t find kernel in '+pk}

    r=ck.load_text_file({'text_file':pk, 'split_to_list':'yes'})
    if r['return']>0: return r

    l=r['lst']

    funcs={}
    desc=[]
    ld=0
    target=''

    cfg=[]
    cfg_start=False

    ll=len(l)
    for k in range(0, ll):
        x=l[k]

        if x.startswith('cfg={'):
           cfg_start=True

        if cfg_start:
           cfg.append(x)
           if x=='    }':
              cfg_start=False

        if x.startswith('##################'):
           desc=[]
           ld=k

        if x.startswith('# '):
           x1=x[2:].strip()
           if x1.startswith('TARGET: '):
              target=x1[8:]
           else:
              desc.append(x1)

        if x.startswith('def '):
           j1=x.find('(')
           j2=x.find(')', j1+1)
           j3=x.find(':')

           fn=x[4:j1]
           i=x[j1+1:j2]
           rem=x[j3+1:]

           api=[]

           k+=1

           found=False
           first=False
           while not found or k<ll:
              x=l[k]
              if x.strip().startswith('"""'):
                 if first:
                    found=True
                    break
                 else:
                    first=True
              elif first:
                 api.append(x)
              k+=1
           
           if found:
              funcs[fn]={'api':api,
                         'input':i,
                         'rem':rem,
                         'desc':desc,
                         'line':ld,
                         'target':target}

           api=[]
           desc=[]
           target=''

    # Process functions
    h+='# CK kernel functions for end-users\n'
    h+='\n'
    h+='We encourage you to reuse these portable productivity functions (Python 2.6+ and 3+) from the CK modules or in your own Python scripts:\n'
    h+='\n'
    h+='```python\n'
    h+='    import ck.kernel as ck\n'
    h+='\n'
    h+='    ck.out("Hello world")\n'
    h+='\n'
    h+='    r=ck.access({"action":"ls", "module_uoa":"env", "out":"con"})\n'
    h+='    if r["return"]>0: ck.err(r)\n'
    h+='    print (r["lst"])\n'
    h+='\n'
    h+='    r=ck.load_json_file({"json_file":"my_file.json", "dict":{"test":"yes"}})\n'
    h+='    if r["return"]>0: return r\n'
    h+='\n'
    h+='    r=ck.save_text_file({"text_file":"my_file.txt", "string":"test=yes\\n"})\n'
    h+='    if r["return"]>0: return r\n'
    h+='\n'
    h+='```\n'

    hdev=h

    num1=0
    num2=0
    for f in sorted(funcs):
        x=funcs[f]

        api=x['api']
        i=x['input']
        rem=x['rem']
        line=x['line']
        desc=x['desc']
        target=x['target']

        xapi='<i>'
        for y in desc:
            j=y.find('\\n=')
            if j>=0: 
               y=y[:j]
            xapi+=y
        xapi+='</i>\n'

        xapi+='\n'
        xapi+='```python\n'
        for y in api:
            xapi+=y+'\n'
        xapi+='```\n'

        url='https://github.com/ctuning/ck/blob/master/ck/kernel.py'

        x=''
        if target!='':
           x+='<p>&nbsp;&nbsp;&nbsp;<i>for '+target+'</i>'

        if 'end users' in target or 'end-users' in target:
           num1+=1
           num=num1
        else:
           num2+=1
           num=num2

        zh='\n## ck.'+f+'('+i+')\n\n'

        zh+=xapi+'\n'
#  <td nowrap valign="top"><a name="'+f+'">'+str(num)+'</td>\n'
#        zh+='  <td nowrap valign="top"><a href="'+url+'#L'+str(line+1)+'"><b>ck.'+f+'('+i+')</b></a>'+x+'</b></td>\n'
#        zh+='  <td nowrap valign="top">'+xapi+'</td>\n'
#        zh+=' </tr>\n'

        if 'end users' in target or 'end-users' in target:
           h+=zh
        else:
           hdev+=zh

    # Prepare config
    hcfg+='You can access the following CK internal variables:\n'
    hcfg+='\n'
    hcfg+='```python\n'
    hcfg+='    import ck.kernel as ck\n'
    hcfg+='\n'
    hcfg+='    print (ck.cfg)\n'
    hcfg+='```\n\n'

    hcfg+='See <a href="https://github.com/ctuning/ck/wiki">CK documentation for further details</a>.\n'

    hcfg+='\n'
    hcfg+='```python\n'
    for x in cfg:
        hcfg+=x+'\n'
    hcfg+='```\n'
    hcfg+=h2

    h+=h2
    hdev+=h2

    if of!='':
       r=ck.save_text_file({'text_file':of+'.md', 'string':h})
       if r['return']>0: return r

       r=ck.save_text_file({'text_file':of+'-dev.md', 'string':hdev})
       if r['return']>0: return r

       r=ck.save_text_file({'text_file':of+'-dev-cfg.md', 'string':hcfg})
       if r['return']>0: return r

    return {'return':0}



#
# Collective Knowledge (CK)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin
#

import sys
import os

##############################################################################


def load_json_file(i):
    """Load json from file into dict
       Target audience: end users

    Args:    
              json_file (str): name of a json file 

    Returns:
              (dict): Unified CK dictionary:

                return (int): return code =  0, if successful
                                          >  0, if error
                (error) (str): error text if return > 0

                dict (dict or list): dict or list from the json file

    """

    import json

    fn = i['json_file']

    try:
        if sys.version_info[0] > 2:
            f = open(fn, 'r', encoding='utf8')
        else:
            f = open(fn, 'r')
    except Exception as e:
        return {'return': 16, 'error': 'problem opening json file='+fn+' ('+format(e)+')'}

    try:
        s = f.read()
    except Exception as e:
        f.close()
        return {'return': 1, 'error': 'problem reading json file='+fn+' ('+format(e)+')'}

    f.close()

    try:
        if sys.version_info[0] > 2:
            d = json.loads(s)
        else:
            d = json.loads(s, encoding='utf8')
    except Exception as e:
        return {'return': 1, 'error': 'problem parsing json from file='+fn+' ('+format(e)+')'}

    return {'return': 0, 'dict': d}

##############################################################################


def save_json_to_file(i):
    """Save dict to a json file
       Target audience: end users

    Args:    
              json_file (str): filename to save dictionary
              dict (dict): dict to save
              (sort_keys) (str): if 'yes', sort keys
              (safe) (str): if 'yes', ignore non-JSON values (only for Debugging - changes original dict!)

    Returns:
              (dict): Unified CK dictionary:

                return (int): return code =  0, if successful
                                          >  0, if error
                (error) (str): error text if return > 0

    """

    import json
    import ck.strings

    fn = i['json_file']

    if i.get('safe', '') == 'yes':
        d = i['dict']

        sd = {}

        # Check main unprintable keys
        for k in d:
            try:
                json.dumps(d[k])
            except Exception as e:
                pass
            else:
                sd[k] = d[k]

        i['dict'] = sd

    r = ck.strings.dump_json(i)
    if r['return'] > 0:
        return r
    s = r['string'].replace('\r', '')+'\n'

    return save_text_file({'text_file': fn, 'string': s})


##############################################################################
def load_yaml_file(i):
    """Load YAML file to dict
       Target audience: end users

    Args:    
              yaml_file (str): name of a YAML file

    Returns:
              (dict): Unified CK dictionary:

                return (int): return code =  0, if successful
                                          >  0, if error
                (error) (str): error text if return > 0

                dict (dict): dict from a YAML file

    """

    import yaml

    fn = i['yaml_file']

    try:
        if sys.version_info[0] > 2:
            f = open(fn, 'r', encoding='utf8')
        else:
            f = open(fn, 'r')
    except Exception as e:
        return {'return': 16, 'error': 'problem opening YAML file='+fn+' ('+format(e)+')'}

    try:
        s = f.read()
    except Exception as e:
        f.close()
        return {'return': 1, 'error': 'problem reading YAML file='+fn+' ('+format(e)+')'}

    f.close()

    try:
        d = yaml.load(s, Loader=yaml.FullLoader)
    except Exception as e:
        return {'return': 1, 'error': 'problem parsing YAML from file='+fn+' ('+format(e)+')'}

    return {'return': 0, 'dict': d}

##############################################################################


def save_yaml_to_file(i):
    """Save dict to a YAML file
       Target audience: end users

    Args:    
              yaml_file (str): name of a YAML file
              dict (dict): dict to save

    Returns:
              (dict): Unified CK dictionary:

                return (int): return code =  0, if successful
                                          >  0, if error
                (error) (str): error text if return > 0

    """

    import yaml

    fn = i['yaml_file']
    d = i['dict']

    try:
        # If using just dump and keys are in unicode,
        # pyyaml adds warning and makes produced yaml unparsable
        s = yaml.safe_dump(d)
    except Exception as e:
        return {'return': 1, 'error': 'problem converting dict to YAML ('+format(e)+')'}

    return save_text_file({'text_file': fn, 'string': s})


##############################################################################
def load_text_file(i):
    """Load a text file to a string or list
       Target audience: end users

    Args:    
              text_file (str): name of a text file
              (keep_as_bin) (str): if 'yes', return only bin
              (encoding) (str): by default 'utf8', however sometimes we use utf16

              (split_to_list) (str): if 'yes', split to list

              (convert_to_dict) (str): if 'yes', split to list and convert to dict
              (str_split) (str): if !='', use as separator of keys/values when converting to dict
              (remove_quotes) (str): if 'yes', remove quotes from values when converting to dict

              (delete_after_read) (str): if 'yes', delete file after read (useful when reading tmp files)

    Returns:
              (dict): Unified CK dictionary:

                return (int): return code =  0, if successful
                                          >  0, if error
                (error) (str): error text if return > 0

                bin (byte): loaded text file as byte array

                (string) (str): loaded text as string with removed \r

                (lst) (list): if split_to_list=='yes', split text to list

                (dict) (dict): if convert_to_dict=='yes', return as dict

    """

    fn = i['text_file']

    en = i.get('encoding', '')
    if en == '' or en == None:
        en = 'utf8'

    try:
        f = open(fn, 'rb')
    except Exception as e:
        return {'return': 16, 'error': 'problem opening text file='+fn+' ('+format(e)+')'}

    try:
        b = f.read()
    except Exception as e:
        f.close()
        return {'return': 1, 'error': 'problem reading text file='+fn+' ('+format(e)+')'}

    f.close()

    r = {'return': 0, 'bin': b}

    if i.get('delete_after_read', '') == 'yes':
        import os
        os.remove(fn)

    if i.get('keep_as_bin', '') != 'yes':
        try:
            # decode into Python string (unicode in Python3)
            s = b.decode(en).replace('\r', '')
        except Exception as e:
            return {'return': 1, 'error': 'problem decoding content from file "'+fn+'" ('+format(e)+')'}

        r['string'] = s

        cl = i.get('split_to_list', '')
        cd = i.get('convert_to_dict', '')

        if cl == 'yes' or cd == 'yes':
            lst = s.split('\n')
            r['lst'] = lst

            if cd == 'yes':
                dd = {}

                ss = i.get('str_split', '')
                rq = i.get('remove_quotes', '')
                if ss == '':
                    ss = ':'

                for q in lst:
                    qq = q.strip()
                    ix = qq.find(ss)
                    if ix > 0:
                        k = qq[0:ix].strip()
                        v = ''
                        if ix+1 < len(qq):
                            v = qq[ix+1:].strip()
                        if v != '' and rq == 'yes':
                            if v.startswith('"'):
                                v = v[1:]
                            if v.endswith('"'):
                                v = v[:-1]
                        dd[k] = v

                r['dict'] = dd

    return r

##############################################################################


def save_text_file(i):
    """Save string to a text file with all \r removed
       Target audience: end users

    Args:    
              text_file (str): name of a text file
              string (str): string to write to a file (all \r will be removed)
              (append) (str): if 'yes', append to a file

    Returns:
              (dict): Unified CK dictionary:

                return (int): return code =  0, if successful
                                          >  0, if error
                (error) (str): error text if return > 0

    """

    fn = i['text_file']

    s = i['string']

    try:
        s = s.replace('\r', '')
    except Exception as e:
        pass

    try:
        s = s.replace(b'\r', b'')
    except Exception as e:
        pass

    m = 'w'
    if i.get('append', '') == 'yes':
        m = 'a'

    try:
        s = s.encode('utf8')
    except Exception as e:
        pass

    try:
        #      if sys.version_info[0]>2:
        #         f=open(fn, m+'b')
        #         f.write(s)
        #      else:
        f = open(fn, m+'b')
        f.write(s)
    except Exception as e:
        return {'return': 1, 'error': 'problem writing text file='+fn+' ('+format(e)+')'}

    f.close()

    return {'return': 0}



#
# Collective Knowledge (description of a given machine for crowd-benchmarking and crowd-tuning)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
line='***************************************************************************************'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# add new machine description

def add(i):
    """
    Input:  {
              (data_uoa) or (alias)        - force name of CK entry to store description of this machine
                                             (if empty, suggest automatically)

              (host_os)                    - host OS (detect, if omitted)
              (target_os)                  - OS module to check (if omitted, analyze host)
              (device_id)                  - device id if remote (such as adb)

              (use_host)                   - if 'yes', configure host as target

              (type) 
                or
              (access_type)                - access type to the machine: 
                                                 "android"
                                                 "mingw"
                                                 "wa_android"
                                                 "wa_linux"
                                                 "ck_node"
                                                 "ssh"
                                                 "rpc"
                                                 "avro"
                                                 "cluster"
                                                 "quantum"

                 Extra options per access type:

                   * avro:
                       avro_config         - full path to JSON configuration file

                   * cluster:
                       cluster_config      - full path to JSON configuration file
                           or
                       config_file

                   * quantum:
                       access_key          - full path to access key file if needed - will be available via env CK_QCK_ACCESS_KEY_FILE
                       user_login          - user login if needed - will be available via env CK_QCK_ACCESS_LOGIN
                       user_password       - user password if needed - will be available via env CK_QCK_ACCESS_PASSWORD

              (share)                      - if 'yes', share public info about platform with the community via cknowledge.org/repo
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import copy
    import os

    # Setting output
    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    # Params
    hos=i.get('host_os','')
    tos=i.get('target_os', '')
    tdid=i.get('device_id', '')

    at=i.get('access_type','')
    if at=='': at=i.get('type','')

    if i.get('use_host','')=='yes':
       at='host'

    duoa=i.get('data_uoa','')
    if duoa=='':
       duoa=i.get('alias','')

    exc='no'
    if i.get('share',''):
        exc='yes'

    er=i.get('exchange_repo','')
    esr=i.get('exchange_subrepo','')

    dp='' # detect platform

    # Preliminary host detect
    os_tags=[]

    r=ck.get_os_ck({})
    if r['return']>0: return r

    if r['platform']=='win':
       os_tags=["windows","mingw"]
    else:
       os_tags=["linux"]

    # If not host target
    tat=cfg['target_access_types']
    if at=='' and o=='con':
        ck.out(line)

        r=ck.select({'title':'Select access type for your target machine: ',
                     'dict':tat})
        if r['return']>0: return r
        at=r['string']

    # If access_type is empty, quit
    if at=='':
        return {'return':1, 'error':'access type is not specified'}

    # Continue processing
    tags=[]
    extra_check={}
    extra_deps={}
    prefix=''
    rtags=''
    if at!='':
       tags=tat[at]['tags']
       dp=tat[at]['detect_platform']
       extra_check=tat[at].get('extra_check',{})
       extra_deps=tat[at].get('extra_deps',{})
       prefix=tat[at].get('alias_prefix','')
       rtags=tat[at].get('record_tags','')
       dos=''

    # Check if use target OS as host OS (useful for Quantum)
    if tat[at].get('use_target_os_as_host','')=='yes':
       ii={'action':'detect',
           'module_uoa':cfg['module_deps']['platform.os'],
           'host_os':hos}
       r=ck.access(ii)
       if r['return']>0: return r
       tos=r['os_uoa']

    # Extra checks
    if tos=='':
        if 'android-with-arch' in tags:
            # Check preliminary Android parameters including arch + API to finalize detecton of the CK OS description
            if o=='con':
                ck.out(line)
                ck.out('Attempting to detect Android API and arch ...')
                ck.out('')

            ii={'action':'detect',
                'host_os':hos,
                'target_os':'android-32',
                'module_uoa':cfg['module_deps']['platform.cpu'],
                'out':oo}
            r=ck.access(ii)
            if r['return']>0: return r

            tdid=r['device_id']

            params=r['features']['os_misc'].get('adb_params',{})

            sdk=str(params.get('ro.build.version.sdk',''))
            abi=str(params.get('ro.product.cpu.abi',''))

            if sdk!='':
                if o=='con':
                   ck.out('')
                   ck.out('Android API: '+sdk)

                for q in range(0, len(tags)):
                    tags[q]+=',android-'+sdk

            if abi!='':
                if o=='con':
                    ck.out('')
                    ck.out('Android ABI: '+abi)

                if abi.startswith('arm64'):
                    dos='*-arm64'
                elif abi.startswith('arm'):
                    dos='*-arm'
                elif abi.startswith('x86_64'):
                    dos='*-x86_64'
                elif abi.startswith('x86'):
                    dos='*-x86'
                elif abi.startswith('mips64'):
                    dos='*-mips64'
                elif abi.startswith('mips'):
                    dos='*-mips'

    # Check tags and use host ones if empty
    if len(tags)==0 and len(os_tags)>0:
        tags=os_tags

    # Search target OS
    if tos=='':
        lst=[]

        for t in tags:
            ii={'action':'search',
                'module_uoa':cfg['module_deps']['os'],
                'data_uoa':dos,
                'tags':t}
            r=ck.access(ii)
            if r['return']>0: return r

            for q in r['lst']:
                lst.append(q)

        if len(lst)==0:
            return {'return':1, 'error':'no OS found for tags "'+str(tags)+'" and OS wildcard "'+dos+'"'}
        elif len(lst)==1:
            tos=lst[0]['data_uoa']
        else:
            ck.out(line)
            ck.out('Select most close OS and architecture on a target machine:')
            ck.out('')

            r=ck.select_uoa({'choices':lst})
            if r['return']>0: return r
            tos=r['choice']

    # Target OS should be finalized
    if tos=='':
        return {'return':1, 'error':'no target OS selected'}

    # Extra deps
    if len(extra_deps)>0:
        ck.out(line)

        r=ck.access({'action':'resolve',
                     'module_uoa':cfg['module_deps']['env'],
                     'deps':extra_deps,
                     'host_os':hos,
                     'target_os':tos,
                     'device_id':tdid,
                     'out':oo})
        if r['return']>0: return r

    # Get user friend alias of OS
    if tos!='':
       r=ck.access({'action':'load',
                    'module_uoa':cfg['module_deps']['os'],
                    'data_uoa':tos})
       if r['return']>0: return r
       tos=r['data_uoa']
       tosd=r['dict']

    if o=='con':
        if tos!='':
            ck.out(line)
            ck.out('Selected target OS UOA:    '+tos)

        if tdid!='':
            ck.out('Selected target device ID: '+tdid)

    # Extra checks if needed
    dd={}
    files={}
    if len(extra_check)>0:
        ii=copy.deepcopy(extra_check)

        ii['out']=oo
        ii['device_id']=tdid

        r=ck.access(ii)
        if r['return']>0: return r

        files=r.get('files',{})
        dd['device_cfg']={'wa_config':r.get('cfg',{})}

    ########################################################## Additional questions for SSH or CK or RPC
    if at=='ssh' or at=='ck_node' or at=='rpc':
        if 'device_cfg' not in dd:
            dd['device_cfg']={}

        dx={}
        henv={}
        keyfile=''

        #####################
        if at=='rpc':
            #####################
            ck.out('')
            r=ck.inp({'text':'Enter hostname (Enter for localhost):          '})

            host=r['string'].strip()
            if host=='': host='localhost'
            dx['host']=host

            #####################
            ck.out('')
            r=ck.inp({'text':'Enter host port if needed:                     '})

            port=r['string'].strip()
            dx['port']=port

        elif at=='ssh':
            #####################
            ck.out('')
            r=ck.inp({'text':'Enter hostname (Enter for localhost):          '})

            host=r['string'].strip()
            if host=='': host='localhost'
            dx['host']=host

            #####################
            ck.out('')
            r=ck.inp({'text':'Enter host port if needed:                     '})

            port=r['string'].strip()
            dx['port']=port

            #####################
            ck.out('')
            r=ck.inp({'text':'Enter username (Enter for root):               '})

            username=r['string'].strip()
            if username=='': username='root'
            dx['username']=username

            #####################
            ck.out('')
            r=ck.inp({'text':'Enter full path to public keyfile:             '})

            keyfile=r['string'].strip()
            dx['keyfile']=keyfile

        else:
            #####################
            ck.out('')
            r=ck.inp({'text':'Enter hostname (Enter for localhost):          '})

            host=r['string'].strip()
            if host=='': host='http://localhost'
            elif not host.startswith('http://'): host='http://'+host
            dx['host']=host

            #####################
            ck.out('')
            r=ck.inp({'text':'Enter host port if needed (Enter for 3333):    '})

            port=r['string'].strip()
            if port=='': port='3333'
            dx['port']=port

            #####################
            ck.out('')
            r=ck.inp({'text':'Enter full local path to file with secret key: '})

            keyfile=r['string'].strip()
            dx['keyfile']=keyfile

        henv['CK_MACHINE_HOST']=dx.get('host','')
        henv['CK_MACHINE_PORT']=dx.get('port','')

        #####################
        path_to_files="tmp-ck"
        if at=='ck_node':
            ck.out('')
            r=ck.inp({'text':'Enter full path to files on a target:          '})

            path_to_files=r['string'].strip()
            dx['path_to_files']=path_to_files

        dd['device_cfg']['remote_params']=dx

        # Prepare OS update
        port1=''
        if port!='':
            port1=':'+port

        keyfile1=''

        if at=='rpc':
           uod={
                 'preset_host_env':henv,
                 'skip_platform_detection':'yes'
               }
        else:
           uod={
            "remote": "yes",
            "remote_ssh":"yes",
             "remote_deinit": "",
             "remote_dir_sep": tosd.get('dir_sep',''),
             "remote_env_quotes_if_space": tosd.get('env_quotes_if_space',''),
             "remote_shell_end": "\""
           }

        if at=='ssh':
            port2=''
            port3=''

            if port!='':
                port2='-p '+port
                port3='-P '+port

            if keyfile!='':
                keyfile1='-i '+keyfile

            uod.update({
                 "remote_dir": path_to_files,
                 "remote_id": username+"@"+host+port1,
                 "remote_init": "ssh "+port2+" -l "+username+" "+host+" "+keyfile1+" \"echo remote_init...\"",
                 "remote_pull": "scp "+port3+" "+keyfile1+" "+username+"@"+host+":$#file1#$ \"$#file2#$\"",
                 "remote_push": "scp "+port3+" "+keyfile1+" \"$#file1#$\" "+username+"@"+host+":$#file2#$",
                 "remote_shell": "ssh "+port2+" -l "+username+" "+host+" "+keyfile1+" \""
                })

        else:
            if keyfile!='':
                keyfile1='--keyfile='+keyfile

            uod.update({
                 "remote_dir": "",
                 "remote_dir_full": path_to_files,
                 "remote_id": host+port1,
                 "remote_init": "ck shell crowdnode --url="+host+port1+" "+keyfile1+" --cmd=\"echo remote_init...\"",
                 "remote_pull": "ck pull crowdnode --url="+host+port1+" "+keyfile1+" --filename=$#file1#$ --filename2=\"$#file2#$\"",
                 "remote_push": "ck push crowdnode --url="+host+port1+" "+keyfile1+" --filename=$#file1#$ --filename2=\"$#file2#$\"",
                 "remote_shell": "ck shell crowdnode --url="+host+port1+" "+keyfile1+" --cmd=\""
                })

        tosd.update(uod)

        dd['device_cfg']['update_target_os_dict']=uod

    ########################################################## Additional questions for Apache AVRO distributed platforms
    if at=='avro':
        if 'device_cfg' not in dd:
            dd['device_cfg']={}

        dx={}
        henv={}

        #####################
        avro_config=i.get('avro_config','')
        if avro_config=='':
           ck.out('')
           r=ck.inp({'text':'Enter path to AVRO JSON file with configuration of all nodes: '})
           avro_config=r['string'].strip()

        # Check if exists
        if not os.path.isfile(avro_config):
           return {'return':1, 'error':'AVRO configuration file not found ('+avro_config+')'}

        # Add to files to save to machine dir
        r=ck.load_json_file({'json_file':avro_config})
        if r['return']>0: return r

        dx['avro_config']=r['dict']

#        henv['CK_MACHINE_HOST']=dx.get('host','')

        r=ck.load_text_file({'text_file':avro_config})
        if r['return']>0: return r
        files['ip']=r['string']

        dd['device_cfg']['remote_params']=dx

        uod={
             'preset_host_env':henv,
             'skip_platform_detection':'yes'
           }

        tosd.update(uod)

        dd['device_cfg']['update_target_os_dict']=uod

    ########################################################## Additional questions for any CLUSTER platform
    if at=='cluster':
        if 'device_cfg' not in dd:
            dd['device_cfg']={}

        dx={}
        henv={}

        #####################
        cluster_config=i.get('cluster_config','')
        if cluster_config=='': cluster_config=i.get('config_file','')
        if cluster_config=='':
           ck.out('')
           r=ck.inp({'text':'Enter path to CLUSTER JSON file with configuration of all nodes: '})
           cluster_config=r['string'].strip()

        # Check if exists
        if not os.path.isfile(cluster_config):
           return {'return':1, 'error':'CLUSTER configuration file not found ('+cluster_config+')'}

        # Add to files to save to machine dir
        r=ck.load_json_file({'json_file':cluster_config})
        if r['return']>0: return r

        dx['cluster_config']=r['dict']

#        henv['CK_MACHINE_HOST']=dx.get('host','')

        r=ck.load_text_file({'text_file':cluster_config})
        if r['return']>0: return r
        files['ip']=r['string']

        dd['device_cfg']['remote_params']=dx

        uod={
             'preset_host_env':henv,
             'skip_platform_detection':'yes'
           }

        tosd.update(uod)

        dd['device_cfg']['update_target_os_dict']=uod

    ########################################################## Additional questions for Quantum platforms
    if at=='quantum':
        if 'device_cfg' not in dd:
            dd['device_cfg']={}

        dx={}
        henv={}

        #####################
        access_key=i.get('access_key','')
        if access_key=='':
           ck.out('')
           r=ck.inp({'text':'Input full path to an access key file for your remote Quantum machine or press Enter if it\'s not used: '})
           access_key=r['string'].strip()

        # Check if exists
        if access_key!='':
           if not os.path.isfile(access_key):
              return {'return':1, 'error':'Access key file was not found ('+access_key+')'}

           dx['qck_access_key_file']=access_key
           henv['CK_QCK_ACCESS_KEY_FILE']=access_key

        #####################
        user_login=i.get('user_login','')
        if user_login=='':
           ck.out('')
           r=ck.inp({'text':'Input user login name or press Enter if it\'s not used: '})
           user_login=r['string'].strip()

        if user_login!='':
           dx['qck_access_login']=user_login
           henv['CK_QCK_ACCESS_LOGIN']=user_login

        #####################
        user_password=i.get('user_password','')
        if user_password=='':
           ck.out('')
           r=ck.inp({'text':'Input user password or press Enter if it\'s not used (note that it will be stored in the CK "machine" entry in a plain text): '})
           user_password=r['string'].strip()

        if user_password!='':
           dx['qck_access_password']=user_password
           henv['CK_QCK_ACCESS_PASSWORD']=user_password

        # Set vars used in program module
        dd['device_cfg']['remote_params']=dx

        uod={
             'preset_host_env':henv,
             'skip_platform_detection':'yes'
           }

        tosd.update(uod)

        dd['device_cfg']['update_target_os_dict']=uod

    # Detect various parameters of the platform (to suggest platform name as well)
    pn=''
    rp={}
    if dp=='yes':
        if o=='con':
            ck.out(line)
            ck.out('Attempting to detect various parameters of your target machine ...')
            ck.out('')

        ii={'action':'detect',
            'host_os':hos,
            'target_os':tos,
            'target_os_dict':tosd,
            'device_id':tdid,
            'module_uoa':cfg['module_deps']['platform'],
            'exchange':exc,
            'exchange_repo':er,
            'exchange_subrepo':esr,
            'out':oo}
        rp=ck.access(ii)
        if rp['return']>0: return rp

        tdid=rp.get('device_id','')

        pn=rp.get('features',{}).get('platform',{}).get('name','')

    # Check if has CPU features for device config
    features=rp.get('features',{})

    xwa=features.get('cpu',{}).get('workload_automation',{}).get('device_config',{})
    if len(xwa)>0:
        if 'device_cfg' not in dd:
            dd['device_cfg']={}
        if 'wa_config' not in dd['device_cfg']:
            dd['device_cfg']['wa_config']={}
        if 'device_config' not in dd['device_cfg']['wa_config']:
            dd['device_cfg']['wa_config']['device_config']={}

        dd['device_cfg']['wa_config']['device_config'].update(xwa)

    # Finalize machine meta
    dd.update({'host_os':hos,
               'target_os':tos,
               'device_id':tdid,
               'host_os_uoa':rp.get('host_os_uoa',''),
               'host_os_uid':rp.get('host_os_uid',''),
               'host_os_dict':rp.get('host_os_dict',{}),
               'target_os_uoa':rp.get('os_uoa',''),
               'target_os_uid':rp.get('os_uid',''),
               'target_os_dict':tosd,
               'access_type':at,
               'features':features})

    # Suggest platform name
    if duoa=='':
        if o=='con':
            ck.out(line)

        if pn!='' and o=='con':
            ck.out('Detected target machine name: '+pn)
            ck.out('')

        if at=='host':
            duoa='host'
        elif pn!='':
            duoa=pn.lower().replace(' ','-').replace('_','-').replace('(','-').replace(')','-').replace('"','-')

        if duoa!='' and prefix!='':
            duoa=prefix+duoa

        if o=='con':
            s='Enter an alias for your machine to be recorded in your CK local repo'

            if duoa!='':
                s+=' or press Enter for "'+duoa+'"'

            s+=' : '

            ck.out('')

            r=ck.inp({'text':s})
            x=r['string'].strip()

            if x!='': duoa=x

    # Check that alias is there
    if duoa=='':
        return {'return':1, 'error':'machine alias is not defined'}

    # Check if entry already exists
    ii={'action':'load',
        'module_uoa':work['self_module_uid'],
        'data_uoa':duoa}
    r=ck.access(ii)
    if r['return']>0 and r['return']!=16:
        return r

    if r['return']==0:
        renew=False

        s='CK entry "machine:'+duoa+'" already exists'
        if o=='con':
            ck.out('')

            r=ck.inp({'text':s+'. Renew (Y/n)? '})
            x=r['string'].strip().lower()

            if x=='' or x=='y' or x=='yes':
                renew=True
            else:
                return {'return':0}

        if not renew:
            return {'return':1, 'error':s}

    ii={'action':'update',
        'module_uoa':work['self_module_uid'],
        'data_uoa':duoa,
        'dict':dd,
        'tags':rtags,
        'substitute':'yes',
        'sort_keys':'yes'}
    r=ck.access(ii)
    if r['return']>0: return r

    duoa=r['data_uoa']
    duid=r['data_uid']

    p=r['path']

    if len(files)>0:
        for f in files:
            xcfg=files[f]
            pp=os.path.join(p,f)

            r=ck.save_text_file({'text_file':pp, 'string':xcfg})
            if r['return']>0: return r

    # Success
    if o=='con':
        ck.out(line)
        ck.out('Your target machine was successfully registered in CK with alias: '+duoa+' ('+duid+')')

    return {'return':0}

##############################################################################
# show registered target machines and their status

def show(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o=i.get('out','')

    if o=='con':
       ck.out(line)
       ck.out('Checking machines ...')
       ck.out('')

    h='<center>\n'

    a=[]

    c='Registered target machines and their status:'

    h+='<h2>'+c+'</h2>\n'

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']
    template=rx['template']

    url=url0
    action=i.get('action','')
    muoa=i.get('module_uoa','')

    st=''

    url+='action=index&module_uoa=wfe&native_action='+action+'&'+'native_module_uoa='+muoa
    url1=url

    # List entries
    r=ck.access({'action':'search',
                 'module_uoa':work['self_module_uid'],
                 'add_meta':'yes'})
    if r['return']>0: return r

    lst=r['lst']

    h+='<table border="1" cellpadding="7" cellspacing="0">\n'

    h+='  <tr>\n'
    h+='   <td align="center"><b>CK alias</b></td>\n'
    h+='   <td align="center"><b>Real machine name</b></td>\n'
    h+='   <td align="center"><b>Native ID</b></td>\n'
    h+='   <td align="center"><b>Target OS name</b></td>\n'
    h+='   <td align="center"><b>OS UOA</b></td>\n'
    h+='   <td align="center"><b>CPUs</b></td>\n'
    h+='   <td align="center"><b>GPUs</b></td>\n'
#    h+='   <td align="center"><b>GPGPUs</b></td>\n'
    h+='   <td align="center"><b>Status</b></td>\n'
    h+='  <tr>\n'

    c+='\n'
    for q in sorted(lst, key = lambda v: v['data_uoa']):
        duoa=q['data_uoa']
        duid=q['data_uid']

        d=q['meta']

        tdid=d.get('device_id','')

        tos=d.get('target_os_uoa','')
        tos_uid=d.get('target_os_uid','')

        r=check({'data_uoa':duoa})
        if r['return']>0: return r

        connected=r['connected']

        at=d.get('access_type','')

        a.append({'data_uoa':duoa, 'data_uid':duid, 'connected':connected})

        # Prepare info
        if connected=='yes':
            ss=' style="background-color:#009f00;color:#ffffff"'
            sx='connected'
        else:
            ss=' style="background-color:#9f0000;color:#ffffff;"'
            sx='not found'

        h+='  <tr>\n'

        ft=d.get('features',{})
        rn=ft.get('platform',{}).get('name','')

        on=ft.get('os',{}).get('name','')

        cpus=''
        for q in ft.get('cpu_unique',[]):
            x=q.get('ck_cpu_name','')
            if x!='':
                if cpus!='': 
                    cpus+='<br>\n'
                cpus+=x

        gpu=ft.get('gpu',{}).get('name','')

        gpgpus=''

        # Prepare HTML
        c+='\n'
        c+=duoa+': '
        h+='   <td align="left"><a href="'+url0+'&wcid='+work['self_module_uid']+':'+duid+'">'+duoa+'</a></td>\n'

        h+='   <td align="center">'+rn+'</td>\n'

        h+='   <td align="center">'+tdid+'</td>\n'

        h+='   <td align="center">'+on+'</td>\n'

        h+='   <td align="center"><a href="'+url0+'&wcid='+cfg['module_deps']['os']+':'+tos_uid+'">'+tos+'</a></td>\n'

        h+='   <td align="center">'+cpus+'</td>\n'

        h+='   <td align="center">'+gpu+'</td>\n'

#        h+='   <td align="center">'+gpgpus+'</td>\n'

        c+=sx
        h+='   <td align="center"'+ss+'>'+sx+'</td>\n'

        h+='  <tr>\n'

    h+='</table>\n'
    h+='</center>\n'

    if o=='con':
       ck.out(line)
       ck.out(c)

    return {'return':0, 'html':h, 'style':st, 'availability':a}

##############################################################################
# view machines in the browser

def browse(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    i['action']='start'
    i['cid']=''
    i['module_uoa']='web'
    i['browser']='yes'
    i['extra_url']='action=index&module_uoa=wfe&native_action=show&native_module_uoa=machine'
    i['template']=''

    return ck.access(i)

##############################################################################
# init machine and update input

def machine_init(i):
    """
    Input:  {
              (target) - target machine
              (input)  - input to update
              (check)  - if 'yes', check status and quit if not connected
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    ii=i.get('input',{})

    target=ii.get('target','')
    if target!='':
        # Check if connected
        if i.get('check','')=='yes':
            r=check({'data_uoa':target})
            if r['return']>0: return r
            if r['connected']!='yes':
               return {'return':1, 'error':'target machine "'+target+'" is not connected'}

        # Load machine entry
        r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':target})
        if r['return']>0: return r

        dd=r['dict']
        pp=r['path'] # can be useful to get access to machine specific files (such as for Apache Avro or CLUSTER IP configuration)

        # Get main parameters
        host_os_uoa=dd.get('host_os_uoa','')
        target_os_uoa=dd.get('target_os_uoa','')
        target_device_id=dd.get('device_id','')

        at=dd.get('access_type','')
        ecfg=dd.get('device_cfg',{})

        ecfg['path_to_ck_target_entry']=pp

        # Update input (if undefined)
        if ii.get('host_os','')=='':
            ii['host_os']=host_os_uoa
        if ii.get('target_os','')=='':
            ii['target_os']=target_os_uoa
        if ii.get('device_id','')=='':
            ii['device_id']=target_device_id
        if len(ii.get('device_cfg',{}))==0:
            ii['device_cfg']=ecfg
            ii['device_cfg']['access_type']=at

    return {'return':0}

##############################################################################
# check machine status (online/offline)

def check(i):
    """
    Input:  {
              data_uoa     - machine UOA
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              connected    - 'yes'/'no'
            }

    """

    o=i.get('out','')

    duoa=i.get('data_uoa','')
    if duoa=='':
       return {'return':1, 'error':'machine UOA is not specified'}

    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duoa})
    if r['return']>0: return r

    d=r['dict']

    device_cfg=d.get('device_cfg',{})

    hos=d.get('host_os_uoa','')

    tos=d.get('target_os_uoa','')
    tos_uid=d.get('target_os_uid','')

    tdid=d.get('device_id','')

    # Check if machine connected
    connected='yes'

    at=d.get('access_type','')

    if at=='host' or at=='wa_linux':
        connected='yes'
    elif at=='rpc':
        import telnetlib

        connected='no'

        rp=device_cfg.get('remote_params',{})
        host=rp.get('host','')
        port=rp.get('port','')

        try:
           tn=telnetlib.Telnet(host, int(port))
           connected='yes'
        except:
           pass

        try:
           tn.write(b"exit\n")
        except:
           pass

        try:
           tn.close
        except:
           pass

    elif at=='avro' or at=='quantum' or at=='cluster':
        connected='yes'

    else:
        # Check status of remote
        connected='no'

        if at=='android' or at=='wa_android' or at=='ssh' or at=='ck_node':
            # Attempt to get Android features 
            ii={'action':'detect',
                'module_uoa':cfg['module_deps']['platform.os'],
                'host_os':hos,
                'target_os':tos,
                'device_cfg':device_cfg,
                'device_id':tdid}
            r=ck.access(ii)

            if r['return']==0:
               connected='yes'

    if o=='con':
       if connected=='yes':
          ck.out('connected')
       else:
          ck.out('offline')

    return {'return':0, 'connected':connected}
